from __future__ import annotations

import json
from abc import ABC, abstractmethod
from ast import unparse
from collections import defaultdict
from collections.abc import Iterable, Iterator
from enum import Enum

from fickling.exception import ResourceExhaustionError
from fickling.fickle import (
    BUILTIN_MODULE_NAMES,
    SAFE_BUILTINS,
    BinGet,
    BinPut,
    Dup,
    Get,
    InterpretationError,
    Interpreter,
    LongBinGet,
    LongBinPut,
    Memoize,
    Pickled,
    Proto,
    Put,
)


class AnalyzerMeta(type):
    _DEFAULT_INSTANCE: Analyzer | None = None

    @property
    def default_instance(cls) -> Analyzer:
        if cls._DEFAULT_INSTANCE is None:
            cls._DEFAULT_INSTANCE = Analyzer(Analysis.ALL)
        return cls._DEFAULT_INSTANCE


class AnalysisContext:
    def __init__(self, pickled: Pickled):
        self.pickled: Pickled = pickled
        self.reported_shortened_code: set[str] = set()
        self.previous_results: list[AnalysisResult] = []
        self.results_by_analysis: dict[type[Analysis], list[AnalysisResult]] = defaultdict(list)

    def analyze(self, analysis: Analysis) -> list[AnalysisResult]:
        try:
            results = list(analysis.analyze(self))
        except ResourceExhaustionError as e:
            # Resource limits exceeded - this is a DoS attack indicator
            results = [
                AnalysisResult(
                    Severity.LIKELY_OVERTLY_MALICIOUS,
                    f"Resource exhaustion detected during analysis: {e}; "
                    f"this is indicative of an expansion attack (Billion Laughs style)",
                    "ResourceExhaustion",
                    trigger=f"{e.resource_type}: {e.actual}",
                )
            ]
        except (ValueError, IndexError, RecursionError) as e:
            # Malformed pickle caused an interpretation error
            results = [
                AnalysisResult(
                    Severity.LIKELY_UNSAFE,
                    f"The pickle file has malformed opcode sequences ({type(e).__name__}: {e}); "
                    f"it is either corrupted or attempting to bypass the pickle security analysis",
                    "InterpretationError",
                )
            ]
        if not results:
            self.results_by_analysis[type(analysis)].append(AnalysisResult(Severity.LIKELY_SAFE))
        else:
            self.previous_results.extend(results)
            self.results_by_analysis[type(analysis)].extend(results)
        return results

    @property
    def results(self) -> AnalysisResults:
        return AnalysisResults(pickled=self.pickled, results=self.previous_results)

    def shorten_code(self, ast_node) -> tuple[str, bool]:
        code = unparse(ast_node).strip()
        if len(code) > 32:
            cutoff = code.find("(")
            if code[cutoff] == "(":
                shortened_code = f"{code[: code.find('(')].strip()}(...)"
            else:
                shortened_code = code
        else:
            shortened_code = code
        was_already_reported = shortened_code in self.reported_shortened_code
        self.reported_shortened_code.add(shortened_code)
        return shortened_code, was_already_reported


class Analyzer(metaclass=AnalyzerMeta):
    def __init__(self, analyses: Iterable[Analysis]):
        self.analyses: tuple[Analysis, ...] = tuple(analyses)

    def analyze(self, pickled: Pickled) -> AnalysisResults:
        context = AnalysisContext(pickled=pickled)
        for a in self.analyses:
            context.analyze(a)
        return context.results


class Severity(Enum):
    LIKELY_SAFE = (0, "No Unsafe Operations Discovered")
    POSSIBLY_UNSAFE = (1, "Possibly Unsafe")
    SUSPICIOUS = (2, "Suspicious")
    LIKELY_UNSAFE = (3, "Likely Unsafe")
    LIKELY_OVERTLY_MALICIOUS = (4, "Likely Overtly Malicious")
    OVERTLY_MALICIOUS = (5, "Overtly Malicious")

    def __lt__(self, other):
        return isinstance(other, Severity) and self.value < other.value

    def __gt__(self, other):
        return not isinstance(other, Severity) or other < self

    def __eq__(self, other):
        return isinstance(other, Severity) and other.value == self.value

    def __ge__(self, other):
        return self > other or self == other

    def __le__(self, other):
        return self < other or self == other

    def __init__(self, severity: int, message: str):
        self.severity: int = severity
        self.message: str = message


class AnalysisResult:
    def __init__(
        self,
        severity: Severity,
        message: str | None = None,
        analysis_name: str | None = None,
        trigger: str | None = None,
    ):
        self.severity: Severity = severity
        self.message: str | None = message
        self.analysis_name: str | None = analysis_name
        self.trigger: str | None = trigger  # Field to store the trigger code fragment or artifact

    def __lt__(self, other):
        return isinstance(other, AnalysisResult) and (
            self.severity < other.severity
            or (self.severity == other.severity and str(self) < str(other))
        )

    def __bool__(self):
        """Returns True if there is no evidence that this result is unsafe"""
        return self.severity == Severity.LIKELY_SAFE

    def __str__(self):
        if self.message is None:
            return "No issues found"
        return self.message


class Analysis(ABC):
    ALL: list[Analysis] = []

    def __init_subclass__(cls, **kwargs):
        Analysis.ALL.append(cls())

    @abstractmethod
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        raise NotImplementedError()


class DuplicateProtoAnalysis(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        had_proto = False
        proto_versions: set[int] = set()
        for i, opcode in enumerate(context.pickled):
            if isinstance(opcode, Proto):
                if had_proto:
                    suffix = self._get_suffix(i)
                    if opcode.version in proto_versions:
                        yield AnalysisResult(
                            Severity.LIKELY_UNSAFE,
                            f"The {i + 1}{suffix} opcode is a duplicate PROTO, which is unusual and may be indicative of a tampered pickle",
                            "DuplicateProtoAnalysis",
                            trigger=str(i + 1),
                        )
                    else:
                        yield AnalysisResult(
                            Severity.LIKELY_UNSAFE,
                            f"The {i + 1}{suffix} opcode is a duplicate PROTO with a different "
                            f"version than reported in the previous PROTO opcode, which is almost "
                            f"certainly a sign of a tampered pickle",
                            "DuplicateProtoAnalysis",
                            trigger=str(i + 1),
                        )
                else:
                    had_proto = True
                proto_versions.add(opcode.version)

    @staticmethod
    def _get_suffix(index: int) -> str:
        return {0: "st", 1: "nd", 2: "rd"}.get(index, "th")


class MisplacedProtoAnalysis(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        for i, opcode in enumerate(context.pickled):
            if isinstance(opcode, Proto):
                if opcode.version >= 2 and i > 0:
                    yield AnalysisResult(
                        Severity.LIKELY_UNSAFE,
                        f"The protocol version is {opcode.version}, but the PROTO opcode is not "
                        f"the first opcode in the pickle, as required for versions 2 and later; "
                        f"this may be indicative of a tampered pickle",
                        "MisplacedProtoAnalysis",
                        trigger=str(opcode.version),
                    )


class InvalidOpcode(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        if context.pickled.has_invalid_opcode:
            yield AnalysisResult(
                Severity.LIKELY_UNSAFE,
                "The file has invalid opcode(s). It is either corrupted or attempting to bypass the pickle security analysis",
            )


class InterpretationErrorAnalysis(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        # Access ast to trigger interpretation and set has_interpretation_error if needed
        _ = context.pickled.ast
        if context.pickled.has_interpretation_error:
            yield AnalysisResult(
                Severity.LIKELY_UNSAFE,
                "The pickle file has malformed opcode sequences. "
                "It is either corrupted or attempting to bypass the pickle security analysis",
            )


class ResourceExhaustionAnalysis(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        if context.pickled.has_resource_exhaustion:
            yield AnalysisResult(
                Severity.LIKELY_OVERTLY_MALICIOUS,
                "Resource limits were exceeded during interpretation; "
                "this is indicative of an expansion attack (Billion Laughs style)",
                "ResourceExhaustion",
            )


class NonStandardImports(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        for node in context.pickled.non_standard_imports():
            shortened, already_reported = context.shorten_code(node)
            if not already_reported:
                yield AnalysisResult(
                    Severity.LIKELY_UNSAFE,
                    f"`{shortened}` imports a Python module that is not a part of "
                    "the standard library; this can execute arbitrary code and is "
                    "inherently unsafe",
                    "NonStandardImports",
                    trigger=shortened,
                )


class UnsafeImportsML(Analysis):
    # ML-specific unsafe modules only; general-purpose modules (os, subprocess,
    # socket, pickle, etc.) are already covered by fickle.py's UNSAFE_IMPORTS
    # frozenset and caught by the UnsafeImports analysis.
    UNSAFE_MODULES = {
        "torch.hub": (
            "This module can load untrusted files from the web, exposing the system to "
            "arbitrary code execution."
        ),
        "torch._dynamo": "This module can compile and execute arbitrary code via JIT.",
        "torch._inductor": "This module can generate and execute compiled code.",
        "torch.jit": "This module can compile and execute arbitrary code.",
        "torch.compile": "This module can compile and execute arbitrary code.",
        "numpy.f2py": "This module can compile and execute arbitrary Fortran/C code.",
        "numpy.distutils": "This module can execute arbitrary build commands.",
    }

    # ML-specific unsafe imports (function-level granularity); general-purpose
    # modules are blocked at the module level by fickle.py's UNSAFE_IMPORTS.
    # _io/io are kept here because blocking the entire io module would flag
    # io.BytesIO which is ubiquitous in ML model loading.
    UNSAFE_IMPORTS = {
        "torch": {
            "load": "This function can load untrusted files and code from arbitrary web sources.",
            "compile": "This function can compile and execute arbitrary code.",
        },
        "torch.storage": {
            "_load_from_bytes": (
                "This function calls `torch.load()` which is unsafe as using a string argument "
                "would allow to load and execute arbitrary code hosted on the internet. However, "
                "in this case, the argument is explicitly converted to `io.bytesIO` and hence "
                "treated as a bytestream and not as a remote URL. However, a malicious file can "
                "supply a pickle opcode bytestring as argument to this function to cause the "
                "underlying `torch.load()` call to unpickle that bytestring and execute arbitrary "
                "code through nested pickle calls. So this import is safe only if restrictions on "
                "pickle (such as Fickling's hooks) have been set properly"
            ),
        },
        "torch.utils._config_module": {
            "ConfigModule": "This class can load configuration which may execute code.",
        },
        "torch.distributed.elastic.rendezvous.api": {
            "basichandlers": "This can execute network handlers.",
        },
        "numpy.testing._private.utils": {"runstring": "This function can execute arbitrary code."},
        "_io": {
            "FileIO": "This class can read/write arbitrary files.",
            "open": "This function can open arbitrary files.",
        },
        "io": {
            "FileIO": "This class can read/write arbitrary files.",
            "open": "This function can open arbitrary files.",
        },
    }

    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        for node in context.pickled.properties.imports:
            shortened, _ = context.shorten_code(node)
            all_modules = [
                node.module.rsplit(".", i)[0] for i in range(0, node.module.count(".") + 1)
            ]
            for module_name in all_modules:
                if module_name in self.UNSAFE_MODULES:
                    # Special handling for builtins - check specific function names
                    if module_name in BUILTIN_MODULE_NAMES:
                        for n in node.names:
                            if n.name not in SAFE_BUILTINS:
                                risk_info = self.UNSAFE_MODULES[module_name]
                                yield AnalysisResult(
                                    Severity.LIKELY_OVERTLY_MALICIOUS,
                                    f"`{shortened}` imports `{n.name}` from `{module_name}` "
                                    f"which can execute arbitrary code. {risk_info}",
                                    "UnsafeImportsML",
                                    trigger=shortened,
                                )
                    else:
                        # All other unsafe modules are fully blocked
                        risk_info = self.UNSAFE_MODULES[module_name]
                        yield AnalysisResult(
                            Severity.LIKELY_OVERTLY_MALICIOUS,
                            f"`{shortened}` uses `{module_name}` that is indicative of a malicious pickle file. {risk_info}",
                            "UnsafeImportsML",
                            trigger=shortened,
                        )
            if node.module in self.UNSAFE_IMPORTS:
                for n in node.names:
                    if n.name in self.UNSAFE_IMPORTS[node.module]:
                        risk_info = self.UNSAFE_IMPORTS[node.module][n.name]
                        yield AnalysisResult(
                            Severity.LIKELY_OVERTLY_MALICIOUS,
                            f"`{shortened}` imports `{n.name}` that is indicative of a malicious pickle file. {risk_info}",
                            "UnsafeImportsML",
                            trigger=shortened,
                        )
            # NOTE(boyan): Special case with eval?
            # Copy pasted from pickled.unsafe_imports() original implementation
            elif "eval" in (n.name for n in node.names):
                yield AnalysisResult(
                    Severity.LIKELY_OVERTLY_MALICIOUS,
                    f"`{shortened}` imports `eval` which can execute arbitrary code",
                    "UnsafeImportsML",
                    trigger=shortened,
                )


class BadCalls(Analysis):
    BAD_CALLS = ["exec", "eval", "compile", "open"]

    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        for node in context.pickled.properties.calls:
            shortened, _already_reported = context.shorten_code(node)
            if any(shortened.startswith(f"{c}(") for c in self.BAD_CALLS):
                yield AnalysisResult(
                    Severity.OVERTLY_MALICIOUS,
                    f"Call to `{shortened}` is almost certainly evidence of a malicious pickle file",
                    "OvertlyBadEval",
                    trigger=shortened,
                )


class OvertlyBadEvals(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        for node in context.pickled.properties.non_setstate_calls:
            if (
                hasattr(node.func, "id")
                and node.func.id in context.pickled.properties.likely_safe_imports
            ):
                # if the call is to a constructor of an object imported from the Python
                # standard library, it's probably okay
                continue
            shortened, already_reported = context.shorten_code(node)
            if (
                shortened.startswith("eval(")
                or shortened.startswith("exec(")
                or shortened.startswith("compile(")
                or shortened.startswith("open(")
                or shortened.startswith("_run_code(")
                or shortened.startswith("execWrapper(")
            ):
                # this is overtly bad, so record it and print it at the end
                yield AnalysisResult(
                    Severity.OVERTLY_MALICIOUS,
                    f"Call to `{shortened}` is almost certainly evidence of a malicious pickle file",
                    "OvertlyBadEval",
                    trigger=shortened,
                )
            elif not already_reported:
                yield AnalysisResult(
                    Severity.LIKELY_UNSAFE,
                    f"Call to `{shortened}` can execute arbitrary code and is inherently unsafe",
                    "OvertlyBadEval",
                    trigger=shortened,
                )


class UnsafeImports(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        for node in context.pickled.unsafe_imports():
            if node.module in BUILTIN_MODULE_NAMES and all(
                n.name in SAFE_BUILTINS for n in node.names
            ):
                continue
            shortened, _ = context.shorten_code(node)
            yield AnalysisResult(
                Severity.LIKELY_OVERTLY_MALICIOUS,
                f"`{shortened}` is suspicious and indicative of an overtly malicious pickle file",
                "UnsafeImports",
                trigger=shortened,
            )


class UnusedVariables(Analysis):
    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        interpreter = Interpreter(context.pickled)
        try:
            unused = interpreter.unused_assignments()
        except (InterpretationError, ResourceExhaustionError):
            # Malformed pickle or resource exhaustion - dedicated analyses will report this
            return
        for varname, asmt in unused.items():
            shortened, _ = context.shorten_code(asmt.value)
            yield AnalysisResult(
                Severity.SUSPICIOUS,
                f"Variable `{varname}` is assigned value `{shortened}` but unused afterward; "
                f"this is suspicious and indicative of a malicious pickle file",
                "UnusedVariables",
                trigger=f"{varname}: {shortened}",
            )


class ExpansionAttackAnalysis(Analysis):
    """Detects potential exponential expansion attacks (Billion Laughs style).

    These attacks use:
    - High GET/PUT ratio: Many GET operations retrieving memoized values
    - Excessive DUP operations: Duplicating stack items repeatedly
    """

    # Thresholds for pattern detection
    DEFAULT_GET_PUT_RATIO_THRESHOLD = 10  # GETs per PUT that is suspicious
    DEFAULT_HIGH_GET_PUT_RATIO_THRESHOLD = 50  # Extremely high ratio
    DEFAULT_DUP_COUNT_THRESHOLD = 100  # Number of DUPs that is suspicious

    def __init__(
        self,
        *,
        get_put_ratio_threshold: int = DEFAULT_GET_PUT_RATIO_THRESHOLD,
        high_get_put_ratio_threshold: int = DEFAULT_HIGH_GET_PUT_RATIO_THRESHOLD,
        dup_count_threshold: int = DEFAULT_DUP_COUNT_THRESHOLD,
    ):
        self._get_put_ratio_threshold = get_put_ratio_threshold
        self._high_get_put_ratio_threshold = high_get_put_ratio_threshold
        self._dup_count_threshold = dup_count_threshold

    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
        get_count = 0
        put_count = 0
        dup_count = 0

        for opcode in context.pickled:
            if isinstance(opcode, BinGet | LongBinGet | Get):
                get_count += 1
            elif isinstance(opcode, BinPut | LongBinPut | Put | Memoize):
                put_count += 1
            elif isinstance(opcode, Dup):
                dup_count += 1

        findings: list[AnalysisResult] = []

        # Check for high GET/PUT ratio
        if put_count > 0:
            ratio = get_count / put_count
            if ratio > self._high_get_put_ratio_threshold:
                findings.append(
                    AnalysisResult(
                        Severity.LIKELY_UNSAFE,
                        f"Extremely high GET/PUT ratio ({ratio:.1f}:1) detected; "
                        f"this pattern is indicative of an exponential expansion attack "
                        f"(Billion Laughs style) that could cause DoS",
                        "ExpansionAttackAnalysis",
                        trigger=f"GET/PUT ratio: {ratio:.1f}:1",
                    )
                )
            elif ratio > self._get_put_ratio_threshold:
                findings.append(
                    AnalysisResult(
                        Severity.SUSPICIOUS,
                        f"High GET/PUT ratio ({ratio:.1f}:1) detected; "
                        f"this may indicate an expansion attack pattern",
                        "ExpansionAttackAnalysis",
                        trigger=f"GET/PUT ratio: {ratio:.1f}:1",
                    )
                )
        elif get_count > self._get_put_ratio_threshold:
            # GETs with no PUTs is inherently malformed/malicious
            findings.append(
                AnalysisResult(
                    Severity.LIKELY_UNSAFE,
                    f"GET operations ({get_count}) with no PUT operations detected; "
                    f"this is indicative of a malformed or malicious pickle",
                    "ExpansionAttackAnalysis",
                    trigger=f"GET count: {get_count}, PUT count: 0",
                )
            )

        # Check for excessive DUP operations
        if dup_count > self._dup_count_threshold:
            findings.append(
                AnalysisResult(
                    Severity.SUSPICIOUS,
                    f"Excessive DUP operations ({dup_count}) detected; "
                    f"this may indicate a stack duplication attack",
                    "ExpansionAttackAnalysis",
                    trigger=f"DUP count: {dup_count}",
                )
            )

        # Multiple indicators together are more severe
        if len(findings) > 1:
            for finding in findings:
                if finding.severity < Severity.LIKELY_UNSAFE:
                    finding.severity = Severity.LIKELY_UNSAFE

        yield from findings


class AnalysisResults:
    def __init__(self, pickled: Pickled, results: Iterable[AnalysisResult]):
        self.pickled: Pickled = pickled
        self.results: tuple[AnalysisResult, ...] = tuple(results)

    @property
    def severity(self) -> Severity:
        if not self.results:
            return Severity.LIKELY_SAFE
        return max(r.severity for r in self.results)

    def __bool__(self):
        """Returns True if all analyses failed to find any unsafe operations"""
        return all(map(bool, sorted(self.results)))

    def detailed_results(self) -> dict[str, dict[str, str]]:
        detailed = defaultdict(dict)
        for result in self.results:
            if result.trigger:
                detailed["AnalysisResult"][result.analysis_name] = result.trigger
        return dict(detailed)

    def to_string(self, verbosity: Severity = Severity.POSSIBLY_UNSAFE):
        return "\n".join(str(r) for r in self.results if verbosity <= r.severity)

    __str__ = to_string

    def to_dict(self, verbosity: Severity = Severity.POSSIBLY_UNSAFE):
        analysis_message = self.to_string(verbosity)
        severity_data = {
            "severity": self.severity.name,
            "analysis": (
                analysis_message
                if analysis_message.strip()
                else "Warning: Fickling failed to detect any overtly unsafe code,"
                "but the pickle file may still be unsafe."
                "Do not unpickle this file if it is from an untrusted source!\n\n"
            ),
            "detailed_results": self.detailed_results(),
        }
        return severity_data


def check_safety(
    pickled: Pickled,
    analyzer: Analyzer | None = None,
    verbosity: Severity = Severity.POSSIBLY_UNSAFE,
    json_output_path: str | None = None,
) -> AnalysisResults:
    if analyzer is None:
        analyzer = Analyzer.default_instance

    results = analyzer.analyze(pickled)

    if json_output_path:
        severity_data = results.to_dict(verbosity)
        # This is intentionally "a" to handle the case of stacked pickles
        with open(json_output_path, "a") as json_file:
            json.dump(severity_data, json_file, indent=4)
    return results


def is_likely_safe(filepath: str):
    with open(filepath, "rb") as f:
        pickled = Pickled.load(f, fail_on_decode_error=False)
        if pickled.has_invalid_opcode:
            # Invalid pickle file, likely not safe
            return False
        return check_safety(pickled).severity == Severity.LIKELY_SAFE



import os
import shutil
import stat
import tarfile
import tempfile
import zipfile
from pathlib import Path
from typing import TypedDict

import numpy.lib.format as npformat

from fickling.fickle import Pickled, StackedPickle

# Optional 7z support
try:
    import py7zr
    import py7zr.exceptions
    from py7zr.io import BytesIOFactory

    HAS_7Z_SUPPORT = True
    _7Z_ARCHIVE_ERROR: type[BaseException] = py7zr.exceptions.ArchiveError
except ImportError:
    HAS_7Z_SUPPORT = False
    _7Z_ARCHIVE_ERROR = Exception  # Fallback, won't be used if HAS_7Z_SUPPORT is False

"""
PyTorch file format identification:

We currently support the following PyTorch file formats:
• PyTorch v0.1.1: Tar file with sys_info, pickle, storages, and tensors
• PyTorch v0.1.10: Stacked pickle files
• TorchScript v1.0: ZIP file with model.json
• TorchScript v1.1: ZIP file with model.json and attributes.pkl (a JSON file and a pickle file)
• TorchScript v1.3: ZIP file with data.pkl and constants.pkl (2 pickle files)
• TorchScript v1.4: ZIP file with data.pkl, constants.pkl, and version set at 2 or higher
• PyTorch v1.3: ZIP file containing data.pkl (1 pickle file)
• PyTorch model archive format[ZIP]: ZIP file that includes Python code files and pickle files

This description draws from this PyTorch GitHub issue: https://github.com/pytorch/pytorch/issues/31877.
If any inaccuracies in that description are found, that should be reflected in this code.
If any new PyTorch file formats are made, that should be added to this code.
Another useful reference is https://github.com/lutzroeder/netron/blob/main/source/pytorch.js.
"""

try:
    from torch.serialization import _is_zipfile
except ModuleNotFoundError:
    raise ImportError(
        "The 'torch' module is required for this functionality."
        "PyTorch is now an optional dependency in Fickling."
        "Please use `pip install fickling[torch]`"
    )


def check_and_find_in_zip(
    zip_path, file_name_or_extension, return_path=False, check_extension=False
):
    """Check for a file in the zip and return its path or boolean if found."""
    try:
        if not (return_path):
            with zipfile.ZipFile(zip_path, "r") as zip_file:
                if check_extension:
                    return any(
                        entry.endswith(file_name_or_extension) for entry in zip_file.namelist()
                    )
                return any(file_name_or_extension in entry for entry in zip_file.namelist())
        else:
            return next(
                (entry for entry in zip_path.namelist() if entry.endswith(file_name_or_extension)),
                None,
            )
    except zipfile.BadZipFile:
        print(f"Invalid ZIP file: {zip_path}")
        return None if return_path else False
    except FileNotFoundError:
        print(f"File not found: {zip_path}")
        return None if return_path else False


def is_7z_file(file_path) -> bool:
    """Check if a file is a 7z archive by reading its magic bytes."""
    try:
        with open(file_path, "rb") as f:
            magic = f.read(6)
            # 7z magic bytes: 37 7A BC AF 27 1C
            return magic == b"7z\xbc\xaf'\x1c"
    except OSError:
        return False


def check_and_find_in_7z(
    archive_path, file_name_or_extension, return_path=False, check_extension=False
):
    """Check for a file in a 7z archive and return its path or boolean if found."""
    if not HAS_7Z_SUPPORT:
        return None if return_path else False

    try:
        with py7zr.SevenZipFile(archive_path, mode="r") as archive:
            names = [f.filename for f in archive.list() if not f.is_symlink]
            if not return_path:
                if check_extension:
                    return any(entry.endswith(file_name_or_extension) for entry in names)
                return any(file_name_or_extension in entry for entry in names)
            return next(
                (entry for entry in names if entry.endswith(file_name_or_extension)),
                None,
            )
    except (OSError, _7Z_ARCHIVE_ERROR):
        return None if return_path else False


def check_numpy(file):  # returns isNumpy,isNumpyPickle
    """Checks if the numpy magic bytes are there, and if they are, if the header
    claims the data is an object"""
    file.seek(0)
    try:
        version = npformat.read_magic(file)
    except ValueError:
        return False, False  # not numpy

    # Use numpy's public API to read the array header
    # This replaces the previous use of private _header_size_info variable
    try:
        if version == (1, 0):
            _shape, _fortran_order, dtype = npformat.read_array_header_1_0(file)
        elif version == (2, 0):
            _shape, _fortran_order, dtype = npformat.read_array_header_2_0(file)
        elif version == (3, 0):
            # Version 3.0 uses the same 4-byte header format as 2.0
            _shape, _fortran_order, dtype = npformat.read_array_header_2_0(file)
        else:
            # Unknown version
            return False, False
    except (ValueError, OSError):
        # Header parsing failed
        return False, False

    if dtype.hasobject:
        return True, True  # numpy pickle
    return True, False  # numpy non-pickle


def check_pickle(file, min_length=0):
    """Checks if a file can be pickled; this does not directly determine the file is a pickle"""
    file.seek(0)
    try:
        p = Pickled.load(file)
        return len(p) > min_length
    except Exception:  # noqa
        file.seek(0)
        try:
            sp = StackedPickle.load(file)
            return any(len(p) > min_length for p in sp.pickled)
        except Exception:  # noqa
            return False


class FileProperties(TypedDict):
    is_torch_zip: bool
    is_tar: bool
    is_valid_pickle: bool
    is_numpy: bool
    is_numpy_pickle: bool
    is_standard_zip: bool
    is_standard_not_torch: bool
    has_constants_pkl: bool
    has_data_pkl: bool
    has_version: bool
    has_model_json: bool
    has_attributes_pkl: bool
    is_7z: bool


def find_file_properties(file_path, print_properties=False) -> FileProperties:
    """For a more granular analysis, we separate property discovery and format identification"""
    properties = {}
    with open(file_path, "rb") as file:
        # PyTorch's torch.load() enforces a specific magic number at offset 0 for ZIP
        is_torch_zip = _is_zipfile(file)
        properties["is_torch_zip"] = is_torch_zip

        # This tarfile check has many false positives. It is not a determinant of PyTorch v0.1.1.
        is_tar = tarfile.is_tarfile(file)
        properties["is_tar"] = is_tar

        # Similar to tar, this is not a robust verification.
        # tar files often start with . which is a 1 operator pickle
        is_valid_pickle = check_pickle(file, min_length=2)
        properties["is_valid_pickle"] = is_valid_pickle

        # Numpy has a special header and magic bytes,
        # mimics Numpy code to check if the file is Numpy
        # and if it claims to contain a pickle
        is_numpy, is_numpy_pickle = check_numpy(file)
        properties["is_numpy"] = is_numpy
        properties["is_numpy_pickle"] = is_numpy_pickle

        # PyTorch MAR can be a standard ZIP, but not a PyTorch ZIP
        # Some other non-PyTorch file formats rely on ZIP without PyTorch's limitations
        is_standard_zip = zipfile.is_zipfile(file)
        properties["is_standard_zip"] = is_standard_zip

        is_standard_not_torch = is_standard_zip and not is_torch_zip
        properties["is_standard_not_torch"] = is_standard_not_torch

        torch_zip_results = {
            "has_constants_pkl": False,
            "has_data_pkl": False,
            "has_version": False,
            "has_model_json": False,
            "has_attributes_pkl": False,
        }
        if is_torch_zip:
            torch_zip_checks = [
                "data.pkl",
                "constants.pkl",
                "version",
                "model.json",
                "attributes.pkl",
            ]
            torch_zip_results = {
                f"has_{'_'.join(f.split('.'))}": check_and_find_in_zip(
                    file, f, check_extension=False
                )
                for f in torch_zip_checks
            }
        properties.update(torch_zip_results)

    # Check for 7z archive (requires file path, not file handle)
    is_7z = is_7z_file(file_path)
    properties["is_7z"] = is_7z

    if print_properties:
        print("\nproperties:", properties, "\n")
    return properties


def find_file_properties_recursively(file_path, print_properties=False):
    """Property discovery that looks inside zip and tar archives"""
    properties = find_file_properties(file_path, print_properties)

    # check zip
    check_zip = properties["is_standard_zip"]
    # if it's the correct type of zip to be torch, make sure it's not a torch file
    if check_zip and not properties["is_standard_not_torch"]:
        for key in properties.keys():
            if key.startswith("has_") and properties[key]:
                # if it is a torch file, no need to check it
                check_zip = False
                break
    # actually check the zip
    if check_zip:
        properties["children"] = {}
        with zipfile.ZipFile(file_path) as zipped_file:
            for fname in zipped_file.namelist():
                info = zipped_file.getinfo(fname)
                mode = info.external_attr >> 16
                if info.is_dir():
                    continue
                if mode == 0 or stat.S_ISREG(mode):
                    _tempfile = tempfile.NamedTemporaryFile(delete=False)
                    try:
                        _tempfile.write(zipped_file.read(fname))
                        _tempfile.close()
                        properties["children"][fname] = find_file_properties_recursively(
                            _tempfile.name, print_properties
                        )
                    finally:
                        Path(_tempfile.name).unlink()

    # check tar
    if properties["is_tar"]:  # tar archive
        properties["children"] = {}
        with tarfile.TarFile(file_path) as tarred_file:
            for fname in tarred_file.getmembers():
                if fname.isfile():
                    if not (content := tarred_file.extractfile(fname)):
                        properties["children"][fname.name] = None
                        continue
                    _tempfile = tempfile.NamedTemporaryFile(delete=False)
                    try:
                        _tempfile.write(content.read())
                        _tempfile.close()
                        properties["children"][fname.name] = find_file_properties_recursively(
                            _tempfile.name, print_properties
                        )
                    finally:
                        Path(_tempfile.name).unlink()

    # check 7z
    if HAS_7Z_SUPPORT and properties.get("is_7z"):
        if "children" not in properties:
            properties["children"] = {}
        try:
            with py7zr.SevenZipFile(file_path, mode="r") as archive:
                entries = archive.list()
                targets = [e.filename for e in entries if e.is_file and not e.is_symlink]
                if targets:
                    max_size = max(e.uncompressed for e in entries if e.filename in set(targets))
                    factory = BytesIOFactory(limit=max_size)
                    archive.extract(targets=targets, factory=factory)
                    for fname in targets:
                        bio = factory.get(fname)
                        bio.seek(0)
                        _tempfile = tempfile.NamedTemporaryFile(delete=False)
                        try:
                            _tempfile.write(bio.read())
                            _tempfile.close()
                            properties["children"][fname] = find_file_properties_recursively(
                                _tempfile.name, print_properties
                            )
                        finally:
                            Path(_tempfile.name).unlink()
        except (OSError, _7Z_ARCHIVE_ERROR):
            # Graceful degradation on 7z errors
            pass

    return properties


def check_if_legacy_format(file):
    """PyTorch v0.1.1: Tar file with sys_info, pickle, storages, and tensors"""
    required_entries = {"pickle", "storages", "tensors"}
    found_entries = set()
    print("check_if_legacy_format")
    try:
        with tarfile.open(file, mode="r:", format=tarfile.PAX_FORMAT) as tar:
            for member in iter(tar.next, None):
                found_entries.add(member.name)
                if required_entries.issubset(found_entries):
                    print("True")
                    return True
    except Exception:  # noqa
        print("False")
        return False


def check_if_model_archive_format(file, properties):
    """
    PyTorch model archive format: ZIP file that includes Python code files and pickle files

    References for the PyTorch Model Archive Format:
    1. https://pytorch.org/serve/getting_started.html
    2. https://github.com/pytorch/serve/tree/master/model-archiver
    """
    if properties["is_standard_zip"]:
        has_json = check_and_find_in_zip(file, ".json", check_extension=True)
        has_serialized_model = check_and_find_in_zip(
            file, ".pt", check_extension=True
        ) or check_and_find_in_zip(file, ".pth", check_extension=True)
        has_code = check_and_find_in_zip(file, ".py", check_extension=True)
        return has_json and has_serialized_model and has_code
    if properties.get("is_7z"):
        has_json = check_and_find_in_7z(file, ".json", check_extension=True)
        has_serialized_model = check_and_find_in_7z(
            file, ".pt", check_extension=True
        ) or check_and_find_in_7z(file, ".pth", check_extension=True)
        has_code = check_and_find_in_7z(file, ".py", check_extension=True)
        return has_json and has_serialized_model and has_code


def check_for_corruption(properties):
    """Checks for corruption at the PyTorch file format level"""
    corrupted = False
    reason = ""
    # We expect this to be expanded upon
    if properties["is_torch_zip"]:
        if (
            properties["has_model_json"]
            and not properties["has_attributes_pkl"]
            and not properties["has_constants_pkl"]
        ):
            corrupted = True
            reason = """Your file may be corrupted. It contained a
            model.json file without an attributes.pkl or constants.pkl file."""
    return corrupted, reason


def identify_pytorch_file_format(file, print_properties=False, print_results=False):
    """
    We are intentionally matching the semantics of the PyTorch reference parsers.
    To be polyglot-aware, we show the file formats ranked by likelihood.
    Our parsing depth is at the file structure level;
    However, it can be at the full parsing level if necessary.
    """
    properties = find_file_properties(file, print_properties)
    formats = []
    corrupted = False
    # The order of this identification is intentional and tries to match PyTorch
    if properties["is_torch_zip"]:
        format_conditions = [
            (["has_data_pkl", "has_constants_pkl", "has_version"], "TorchScript v1.4"),
            (["has_data_pkl", "has_constants_pkl"], "TorchScript v1.3"),
            (["has_model_json", "has_constants_pkl"], "TorchScript v1.0"),
            (["has_model_json", "has_attributes_pkl"], "TorchScript v1.1"),
            (["has_data_pkl"], "PyTorch v1.3"),
        ]
        formats = [
            format_name
            for keys, format_name in format_conditions
            if all(properties[key] for key in keys)
        ]

    if properties["is_tar"]:
        is_pytorch_legacy_format = check_if_legacy_format(file)
        if is_pytorch_legacy_format:
            formats.append("PyTorch v0.1.1")
    if properties["is_valid_pickle"]:
        formats.append("PyTorch v0.1.10")
    if properties["is_standard_zip"] or properties.get("is_7z"):
        is_model_archive_format = check_if_model_archive_format(file, properties)
        if is_model_archive_format:
            formats.append("PyTorch model archive format")
    corrupted, reason = check_for_corruption(properties)

    # Show results to user
    if corrupted:
        print(reason)
    if len(formats) != 0:
        primary = formats[0]
        if print_results:
            print("Your file is most likely of this format: ", primary, "\n")
        secondary = formats[1:]
        if len(secondary) != 0:
            if print_results:
                print(
                    "It is also possible that your file can be validly interpreted as: ", secondary
                )
    else:
        if print_results:
            print(
                """Your file may not be a PyTorch file.
                No valid file formats were detected.
                If this is a mistake, raise an issue on our GitHub."""
            )
    return formats


def append_file(source_filename, destination_filename):
    with open(source_filename, "rb") as source_file:
        content = source_file.read()
    with open(destination_filename, "ab") as destination_file:
        destination_file.write(content)
    return destination_filename


def create_zip_pickle_polyglot(zip_file, pickle_file):
    append_file(zip_file, pickle_file)


def create_mar_legacy_pickle_polyglot(
    files, print_results=False, polyglot_file_name="polyglot.mar.pt"
):
    files.sort(key=lambda x: x[1] != "PyTorch model archive format")
    if print_results:
        print("Making a PyTorch MAR/PyTorch v0.1.10 polyglot")
    polyglot_file = append_file(*[file[0] for file in files])
    shutil.copy(polyglot_file, polyglot_file_name)
    polyglot_found = True
    return polyglot_found


def create_standard_torchscript_polyglot(
    files, print_results=False, polyglot_file_name="polyglot.pt"
):
    if print_results:
        print("Making a PyTorch v1.3/TorchScript v1.4 polyglot")
        print("Warning: For some parsers, this may generate polymocks instead of polyglots.")
    standard_pytorch_file = [file[0] for file in files if file[1] == "PyTorch v1.3"][0]
    torchscript_file = [file[0] for file in files if file[1] == "TorchScript v1.4"][0]
    if polyglot_file_name is None:
        polyglot_file_name = "polyglot.pt"
    shutil.copy(standard_pytorch_file, polyglot_file_name)

    with zipfile.ZipFile(torchscript_file, "r") as zip_b:
        constants_pkl_path = check_and_find_in_zip(
            zip_b, "constants.pkl", check_extension=False, return_path=True
        )
        version_path = check_and_find_in_zip(zip_b, "version", return_path=True)
        if constants_pkl_path and version_path:
            constants_data = zip_b.read(constants_pkl_path)
            version_data = zip_b.read(version_path)

            with zipfile.ZipFile(polyglot_file_name, "a") as zip_out:
                zip_out.writestr("constants.pkl", constants_data)
                zip_out.writestr("version", version_data)

    polyglot_found = True
    return polyglot_found


def create_mar_legacy_tar_polyglot(
    files, print_results=False, polyglot_file_name="polyglot.mar.tar"
):
    if print_results:
        print("Making a PyTorch v0.1.1/PyTorch MAR polyglot")
    mar_file = [file[0] for file in files if file[1] == "PyTorch model archive format"][0]
    tar_file = [file[0] for file in files if file[1] == "PyTorch v0.1.1"][0]
    polyglot_file = append_file(mar_file, tar_file)
    shutil.copy(polyglot_file, polyglot_file_name)
    polyglot_found = True
    return polyglot_found


def create_polyglot(first_file, second_file, polyglot_file_name=None, print_results=True):
    polyglot_found = False
    temp_first_file = "temp_" + os.path.basename(first_file)
    temp_second_file = "temp_" + os.path.basename(second_file)
    shutil.copy(first_file, temp_first_file)
    shutil.copy(second_file, temp_second_file)
    files = [
        (temp_first_file, identify_pytorch_file_format(temp_first_file)[0]),
        (temp_second_file, identify_pytorch_file_format(temp_second_file)[0]),
    ]
    formats = set(map(lambda x: x[1], files))  # noqa
    if {"PyTorch model archive format", "PyTorch v0.1.10"}.issubset(formats):
        if polyglot_file_name is None:
            polyglot_file_name = "polyglot.mar.pt"
        polyglot_found = create_mar_legacy_pickle_polyglot(files, print_results, polyglot_file_name)
    if {"PyTorch v1.3", "TorchScript v1.4"}.issubset(formats):
        if polyglot_file_name is None:
            polyglot_file_name = "polyglot.pt"
        polyglot_found = create_standard_torchscript_polyglot(
            files, print_results, polyglot_file_name
        )
    if {"PyTorch model archive format", "PyTorch v0.1.1"}.issubset(formats):
        if polyglot_file_name is None:
            polyglot_file_name = "polyglot.mar.tar"
        polyglot_found = create_mar_legacy_tar_polyglot(files, print_results, polyglot_file_name)
    if print_results:
        if polyglot_found is False:
            print(
                """Fickling was not able to create any polyglots.
                  If you think this is a mistake, raise an issue on our GitHub."""
            )
        else:
            print(f"The polyglot is contained in {polyglot_file_name}")
    os.remove(temp_first_file)
    os.remove(temp_second_file)
    return polyglot_found



import _pickle
import io
import pickle

import fickling.loader as loader
from fickling.ml import FicklingMLUnpickler

_original_pickle_load = pickle.load
_original_pickle_loads = pickle.loads
_original_pickle_Unpickler = pickle.Unpickler
_original__pickle_Unpickler = _pickle.Unpickler


class FicklingSafetyUnpickler:
    """
    Drop-in replacement for pickle.Unpickler that uses fickling's safety analysis.

    This class intercepts direct uses of pickle.Unpickler() (e.g., by PyTorch v1.3+)
    and routes them through fickling's load() function for security analysis.
    """

    def __init__(self, file, *args, **kwargs):
        self._file = file
        self._args = args
        self._kwargs = kwargs

    def load(self):
        """Delegate to fickling.load() for security analysis"""
        return loader.load(self._file, *self._args, **self._kwargs)


def run_hook():
    """Replace pickle.load() and pickle.Unpickler by fickling's safe versions"""
    # Hook functions
    pickle.load = loader.load
    _pickle.load = loader.load
    pickle.loads = loader.loads
    _pickle.loads = loader.loads

    # Hook the Unpickler class
    pickle.Unpickler = FicklingSafetyUnpickler
    _pickle.Unpickler = FicklingSafetyUnpickler


def always_check_safety():
    """
    Alias for run_hook()
    """
    run_hook()


def activate_safe_ml_environment(also_allow=None):
    """Enforce using the ML whitelist unpickler"""

    def new_load(file, *args, **kwargs):
        return FicklingMLUnpickler(file, also_allow=also_allow, **kwargs).load(*args)

    def new_loads(data, *args, **kwargs):
        return FicklingMLUnpickler(io.BytesIO(data), also_allow=also_allow, **kwargs).load(*args)

    # Hook functions
    pickle.load = new_load
    _pickle.load = new_load
    pickle.loads = new_loads
    _pickle.loads = new_loads

    # Hook Unpickler class - create a subclass that passes also_allow
    class SafeMLUnpickler(FicklingMLUnpickler):
        """Unpickler with pre-configured also_allow list"""

        def __init__(self, file, *args, **kwargs):
            super().__init__(file, *args, also_allow=also_allow, **kwargs)

    pickle.Unpickler = SafeMLUnpickler
    _pickle.Unpickler = SafeMLUnpickler


def remove_hook():
    """Restore original pickle functions and classes"""
    pickle.load = _original_pickle_load
    _pickle.load = _original_pickle_load
    pickle.loads = _original_pickle_loads
    _pickle.loads = _original_pickle_loads
    pickle.Unpickler = _original_pickle_Unpickler
    _pickle.Unpickler = _original__pickle_Unpickler


# Alias
deactivate_safe_ml_environment = remove_hook



import ast
from ast import unparse

from .fickle import Interpreter, MarkObject, Opcode, Stack


class Trace:
    def __init__(self, interpreter: Interpreter):
        self.interpreter: Interpreter = interpreter

    def on_pop(self, popped_value: ast.expr | MarkObject):
        if isinstance(popped_value, MarkObject):
            value = "MARK"
        else:
            value = unparse(popped_value).strip()
        print(f"\tPopped {value}")

    def on_push(self, pushed_value: ast.expr | MarkObject):
        if isinstance(pushed_value, MarkObject):
            value = "MARK"
        else:
            value = unparse(pushed_value).strip()
        print(f"\tPushed {value}")

    def on_memoize(self, index: int, value: ast.expr):
        print(f"\tMemoized {index} -> {unparse(value).strip()}")

    def on_update_memo(self, index: int, old_value: ast.expr, new_value: ast.expr):
        print(
            f"\tMemo index {index} changed from {unparse(old_value).strip()} to "
            f"{unparse(new_value).strip()}"
        )

    def on_statement(self, statement: ast.stmt):
        print(f"\t{unparse(statement).strip()}")

    def on_opcode(self, opcode: Opcode):
        print(opcode.name)

    def run(self) -> ast.AST:
        while True:
            memory_before = dict(self.interpreter.memory)
            stack_before = Stack(self.interpreter.stack)
            len_module_before = len(self.interpreter.module_body)
            try:
                opcode = self.interpreter.step()
            except StopIteration:
                break
            self.on_opcode(opcode)
            for added in self.interpreter.module_body[len_module_before:]:
                self.on_statement(added)
            common_prefix_length = 0
            for before, after in zip(stack_before, self.interpreter.stack, strict=False):
                if before != after:
                    break
                common_prefix_length += 1
            for before in reversed(stack_before[common_prefix_length:]):
                self.on_pop(before)
            for after in self.interpreter.stack[common_prefix_length:]:
                self.on_push(after)
            for k, v in self.interpreter.memory.items():
                if k not in memory_before:
                    self.on_memoize(k, v)
                elif v != memory_before[k]:
                    self.on_update_memo(k, memory_before[k], v)
        return self.interpreter.to_ast()



from __future__ import annotations

import sys
from argparse import ArgumentParser
from ast import unparse
from pathlib import PurePosixPath

from . import __version__, fickle, tracing
from .analysis import Severity, check_safety
from .constants import EXIT_CLEAN, EXIT_ERROR, EXIT_UNSAFE
from .exception import ResourceExhaustionError
from .loader import scan_file, scan_zip_archive

DEFAULT_JSON_OUTPUT_FILE = "safety_results.json"

HF_RAW_PICKLE_EXTENSIONS = frozenset({".bin", ".pkl", ".pickle"})
HF_ZIP_PICKLE_EXTENSIONS = frozenset({".pt", ".pth"})
HF_PICKLE_EXTENSIONS = HF_RAW_PICKLE_EXTENSIONS | HF_ZIP_PICKLE_EXTENSIONS
HF_KNOWN_SAFE_EXTENSIONS = frozenset({".json", ".safetensors", ".onnx"})
HF_KNOWN_SAFE_FILES = frozenset({".gitattributes", ".gitignore", "README.md"})


def _scan_huggingface(
    repo_id: str,
    revision: str | None = None,
    token: str | None = None,
    json_output_path: str | None = None,
    print_results: bool = False,
) -> int:
    """Scan a HuggingFace Hub repository for potentially malicious pickle files.

    Args:
        repo_id: HuggingFace repository ID (e.g., 'bert-base-uncased')
        revision: Specific revision (branch, tag, or commit) to scan
        token: HuggingFace API token for private repositories
        json_output_path: Path to write JSON results
        print_results: Whether to print results to console

    Returns:
        EXIT_CLEAN (0), EXIT_UNSAFE (1), or EXIT_ERROR (2)
    """
    try:
        from huggingface_hub import HfApi, hf_hub_download

        from .polyglot import find_file_properties
    except ImportError:
        sys.stderr.write(
            "Error: huggingface_hub and torch are required for --huggingface scanning.\n"
            "Install with: pip install fickling[huggingface]\n"
        )
        return EXIT_ERROR

    api = HfApi(token=token)

    try:
        repo_info = api.repo_info(repo_id=repo_id, revision=revision, token=token)
    except Exception as e:  # noqa: BLE001 -- HF Hub may raise unpredictable HTTP errors
        sys.stderr.write(f"Error accessing HuggingFace repository '{repo_id}': {e!s}\n")
        return EXIT_ERROR

    if repo_info is None or not repo_info.siblings:
        if print_results:
            print(f"No files found in {repo_id}")
        return EXIT_CLEAN

    files_to_scan = [f.rfilename for f in repo_info.siblings]

    if not files_to_scan:
        if print_results:
            print(f"No scannable files found in {repo_id}")
        return EXIT_CLEAN

    if print_results:
        print(f"Scanning {len(files_to_scan)} file(s) in {repo_id}...")

    overall_safe = True
    failed = 0
    json_output = json_output_path or DEFAULT_JSON_OUTPUT_FILE

    for filename in files_to_scan:
        try:
            local_path = hf_hub_download(
                repo_id=repo_id,
                filename=filename,
                revision=revision,
                token=token,
            )
        except Exception as e:  # noqa: BLE001 -- HF Hub may raise unpredictable errors
            sys.stderr.write(f"Error downloading {filename}: {e!s}\n")
            overall_safe = False
            failed += 1
            continue

        ext = PurePosixPath(filename).suffix.lower()
        props = find_file_properties(local_path)
        is_zip = (
            props["is_torch_zip"] or props["is_standard_zip"] or ext in HF_ZIP_PICKLE_EXTENSIONS
        )
        if is_zip:
            if print_results:
                print(f"\n  Scanning: {filename}")
            member_results = scan_zip_archive(
                local_path, graceful=True, json_output_path=json_output
            )
            file_results = list(member_results.values())
        elif props["is_valid_pickle"] or ext in HF_RAW_PICKLE_EXTENSIONS:
            if print_results:
                print(f"\n  Scanning: {filename}")
            file_results = [scan_file(local_path, graceful=True, json_output_path=json_output)]
        elif props.get("is_7z") or props["is_tar"] or props["is_numpy_pickle"]:
            sys.stderr.write(
                f"Warning: '{filename}' detected as 7z/TAR/NumPy archive but scanning "
                f"these formats is not yet supported; skipping\n"
            )
            continue
        else:
            name = PurePosixPath(filename).name
            if ext not in HF_KNOWN_SAFE_EXTENSIONS and name not in HF_KNOWN_SAFE_FILES:
                sys.stderr.write(f"Warning: skipping '{filename}' (unknown extension '{ext}')\n")
            elif print_results:
                print(f"  Skipping safe file: {filename}")
            continue

        for result in file_results:
            if print_results:
                for ar in result.results:
                    result_str = ar.to_string()
                    if result_str:
                        print(f"    {result_str}")

            if not result.is_safe:
                overall_safe = False
                if print_results:
                    sys.stderr.write(f"    WARNING: {filename} may contain unsafe content!\n")

            if result.errors:
                for err in result.errors:
                    sys.stderr.write(f"    {err}\n")
                failed += 1

    if print_results:
        if failed > 0:
            sys.stderr.write(f"\nWARNING: {failed}/{len(files_to_scan)} file(s) failed to scan\n")
        if overall_safe:
            print(f"\n{repo_id}: No obvious safety issues detected")
        else:
            print(f"\n{repo_id}: Potentially unsafe content detected!")

    return EXIT_CLEAN if overall_safe else EXIT_UNSAFE


def main(argv: list[str] | None = None) -> int:
    if argv is None:
        argv = sys.argv

    parser = ArgumentParser(
        description="fickling is a static analyzer and interpreter for Python pickle data"
    )
    parser.add_argument(
        "PICKLE_FILE",
        type=str,
        nargs="?",
        default="-",
        help="path to the pickle file to either "
        "analyze or create (default is '-' for "
        "STDIN/STDOUT)",
    )
    options = parser.add_mutually_exclusive_group()
    options.add_argument(
        "--inject",
        "-i",
        type=str,
        default=None,
        help="inject the specified Python code to be run at the end of unpickling, "
        "and output the resulting pickle data",
    )
    parser.add_argument(
        "--inject-target",
        type=int,
        default=0,
        help="some machine learning frameworks stack multiple pickles into the same model file; "
        "this option specifies the index of the pickle file in which to inject the code from the "
        "`--inject` command (default is 0)",
    )
    options.add_argument("--create", "-c", type=str, default=None)
    parser.add_argument(
        "--run-last",
        "-l",
        action="store_true",
        help="used with --inject to have the injected code "
        "run after the existing pickling code in "
        "PICKLE_FILE (default is for the injected code "
        "to be run before the existing code)",
    )
    parser.add_argument(
        "--replace-result",
        "-r",
        action="store_true",
        help=(
            "used with --inject to replace the unpickling result of the code in PICKLE_FILE "
            "with the return value of the injected code. Either way, the preexisting pickling "
            "code is still executed."
        ),
    )
    options.add_argument(
        "--check-safety",
        "-s",
        action="store_true",
        help=(
            "test if the given pickle file is known to be unsafe. If so, exit with non-zero "
            "status. This test is not guaranteed correct; the pickle file may still be unsafe "
            "even if this check exits with code zero."
        ),
    )

    parser.add_argument(
        "--json-output",
        type=str,
        default=None,
        help="path to the output JSON file to store the analysis results from check-safety."
        f"If not provided, a default file named {DEFAULT_JSON_OUTPUT_FILE} will be used.",
    )

    parser.add_argument(
        "--print-results",
        "-p",
        action="store_true",
        help="Print the analysis results to the console when checking safety.",
    )

    parser.add_argument(
        "--trace",
        "-t",
        action="store_true",
        help="print a runtime trace while interpreting the input pickle file",
    )
    parser.add_argument("--version", "-v", action="store_true", help="print the version and exit")
    options.add_argument(
        "--huggingface",
        "--hf",
        type=str,
        default=None,
        metavar="REPO_ID",
        help="scan a model from HuggingFace Hub by repository ID (e.g., 'bert-base-uncased'). "
        "Requires huggingface_hub: pip install fickling[huggingface]",
    )
    parser.add_argument(
        "--hf-revision",
        type=str,
        default=None,
        help="specific revision (branch, tag, or commit) to scan from HuggingFace Hub",
    )
    parser.add_argument(
        "--hf-token",
        type=str,
        default=None,
        help="HuggingFace API token for private repositories. "
        "Prefer setting the HF_TOKEN environment variable to avoid token exposure in "
        "process listings",
    )

    args = parser.parse_args(argv[1:])

    if args.version:
        if sys.stdout.isatty():
            print(f"fickling version {__version__}")
        else:
            print(__version__)
        return EXIT_CLEAN

    # HuggingFace scanning mode
    if args.huggingface is not None:
        return _scan_huggingface(
            repo_id=args.huggingface,
            revision=args.hf_revision,
            token=args.hf_token,
            json_output_path=args.json_output,
            print_results=args.print_results,
        )

    if args.create is None:
        if args.PICKLE_FILE == "-":
            if hasattr(sys.stdin, "buffer") and sys.stdin.buffer is not None:
                file = sys.stdin.buffer
            else:
                file = sys.stdin
        else:
            file = open(args.PICKLE_FILE, "rb")
        try:
            stacked_pickled = fickle.StackedPickle.load(file, fail_on_decode_error=False)
        except fickle.PickleDecodeError as e:
            sys.stderr.write(f"Fickling failed to parse this pickle file. Error: {e!s}\n")
            if args.check_safety:
                sys.stderr.write(
                    "Parsing errors might be indicative of a maliciously crafted pickle file. DO NOT TRUST this file without performing further analysis!\n"
                )
                sys.stderr.write(
                    "\n(If this is a valid pickle file, please report the error at https://github.com/trailofbits/fickling)\n"
                )
            return EXIT_ERROR
        finally:
            file.close()

        if args.inject is not None:
            if args.inject_target >= len(stacked_pickled):
                sys.stderr.write(
                    f"Error: --inject-target {args.inject_target} is too high; there are only "
                    f"{len(stacked_pickled)} stacked pickle files in the input\n"
                )
                return EXIT_ERROR
            if hasattr(sys.stdout, "buffer") and sys.stdout.buffer is not None:
                buffer = sys.stdout.buffer
            else:
                buffer = sys.stdout
            for pickled in stacked_pickled[: args.inject_target]:
                pickled.dump(buffer)
            pickled = stacked_pickled[args.inject_target]
            if not isinstance(pickled[-1], fickle.Stop):
                sys.stderr.write(
                    "Warning: The last opcode of the input file was expected to be STOP, but was "
                    f"in fact {pickled[-1].info.name}"
                )
            pickled.insert_python_eval(
                args.inject,
                run_first=not args.run_last,
                use_output_as_unpickle_result=args.replace_result,
            )
            pickled.dump(buffer)
            for pickled in stacked_pickled[args.inject_target + 1 :]:
                pickled.dump(buffer)
        elif args.check_safety:
            was_safe = True
            json_output_path = args.json_output or DEFAULT_JSON_OUTPUT_FILE
            for pickled in stacked_pickled:
                safety_results = check_safety(pickled, json_output_path=json_output_path)

                # Print results if requested
                if args.print_results:
                    print(safety_results.to_string())

                if safety_results.severity > Severity.LIKELY_SAFE:
                    was_safe = False
                    if args.print_results:
                        sys.stderr.write(
                            "Warning: Fickling detected that the pickle file may be unsafe.\n\n"
                            "Do not unpickle this file if it is from an untrusted source!\n\n"
                        )

            return EXIT_CLEAN if was_safe else EXIT_UNSAFE

        else:
            var_id = 0
            for i, pickled in enumerate(stacked_pickled):
                interpreter = fickle.Interpreter(
                    pickled, first_variable_id=var_id, result_variable=f"result{i}"
                )
                try:
                    if args.trace:
                        trace = tracing.Trace(interpreter)
                        print(unparse(trace.run()))
                    else:
                        print(unparse(interpreter.to_ast()))
                except ResourceExhaustionError as e:
                    sys.stderr.write(
                        f"Error: {e}\n"
                        "This pickle file may contain an expansion attack. "
                        "Use --check-safety to analyze it.\n"
                    )
                    return 1
                var_id = interpreter.next_variable_id
    else:
        pickled = fickle.Pickled(
            [
                fickle.Global.create("__builtin__", "eval"),
                fickle.Mark(),
                fickle.Unicode(args.create.encode("utf-8")),
                fickle.Tuple(),
                fickle.Reduce(),
                fickle.Stop(),
            ]
        )
        if args.PICKLE_FILE == "-":
            file = sys.stdout
            if hasattr(file, "buffer") and file.buffer is not None:
                file = file.buffer
        else:
            file = open(args.PICKLE_FILE, "wb")
        try:
            pickled.dump(file)
        finally:
            file.close()

    return EXIT_CLEAN



class UnsafeFileError(Exception):
    def __init__(self, filepath, info):
        super().__init__()
        self.filepath = filepath
        self.info = info

    def __str__(self):
        return f"Safety results for {self.filepath} : {self.info!s}"


class WrongMethodError(Exception):
    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return self.msg


class ResourceExhaustionError(Exception):
    """Raised when resource limits are exceeded during analysis."""

    def __init__(self, resource_type: str, limit: int, actual: int):
        self.resource_type = resource_type
        self.limit = limit
        self.actual = actual
        super().__init__(
            f"Resource limit exceeded: {resource_type} (limit={limit}, actual={actual})"
        )


class ExpansionAttackError(ResourceExhaustionError):
    """Raised when exponential expansion attack (Billion Laughs style) is detected."""

    def __init__(self, limit: int, actual: int):
        super().__init__("get_ratio", limit, actual)



from __future__ import annotations

import ast
import keyword
import marshal
import re
import struct
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator, MutableSequence, Sequence
from dataclasses import dataclass
from enum import Enum
from io import BytesIO
from pickletools import OpcodeInfo, genops, opcodes
from typing import (
    Any,
    BinaryIO,
    Generic,
    TypeVar,
    overload,
)

from fickling.exception import ExpansionAttackError, ResourceExhaustionError, WrongMethodError

T = TypeVar("T")

if sys.version_info < (3, 12):
    from typing_extensions import Buffer
else:
    from collections.abc import Buffer


@dataclass(frozen=True)
class InterpreterLimits:
    """Resource limits to prevent DoS attacks during pickle interpretation."""

    max_opcodes: int = 1_000_000
    max_stack_depth: int = 10_000
    max_memo_size: int = 100_000
    max_get_ratio: int = 50  # Maximum GETs per PUT threshold

    def __post_init__(self) -> None:
        for field_name in ("max_opcodes", "max_stack_depth", "max_memo_size", "max_get_ratio"):
            value = getattr(self, field_name)
            if value < 1:
                raise ValueError(f"{field_name} must be positive, got {value}")


# Default limits instance (frozen, so safe as a global singleton)
DEFAULT_INTERPRETER_LIMITS = InterpreterLimits()


OpcodeSequence = MutableSequence["Opcode"]
GenericSequence = Sequence[T]
make_constant = ast.Constant

BUILTIN_STDLIB_MODULE_NAMES: frozenset[str] = sys.stdlib_module_names

OPCODES_BY_NAME: dict[str, type[Opcode]] = {}
OPCODE_INFO_BY_NAME: dict[str, OpcodeInfo] = {opcode.name: opcode for opcode in opcodes}

UNSAFE_IMPORTS: frozenset[str] = frozenset(
    [
        # Builtins - can execute arbitrary code
        "__builtin__",
        "__builtins__",
        "builtins",
        # System/process execution
        "os",
        "posix",
        "nt",
        "subprocess",
        "sys",
        "pty",
        "commands",  # Legacy Python 2 module
        "multiprocessing",
        "_posixsubprocess",
        # Code execution/compilation
        "code",
        "codeop",
        "runpy",
        "marshal",
        "types",
        "compile",
        "exec",
        "eval",
        "doctest",
        # Import manipulation
        "importlib",
        "_frozen_importlib",
        "_frozen_importlib_external",
        "_imp",
        "pkgutil",
        "zipimport",
        "gc",
        "site",
        "unittest",
        # Attribute access (getattr equivalent bypasses)
        "inspect",
        "test",
        # Operator module bypasses
        "_operator",
        "operator",
        "functools",
        # Profiling/debugging (can execute code)
        "cProfile",
        "profile",
        "pdb",
        "bdb",
        "timeit",
        "trace",
        # Network - data exfiltration/download
        "socket",
        "ssl",
        "httplib",
        "http",
        "urllib",
        "urllib2",
        "requests",
        "aiohttp",
        "asyncio",  # Can run arbitrary coroutines
        "webbrowser",  # Can open arbitrary URLs
        "smtplib",
        "imaplib",
        "ftplib",
        "poplib",
        "telnetlib",
        "nntplib",
        # FFI/native code execution
        "ctypes",
        "_ctypes",
        # Pickle recursion (nested pickle attacks)
        "pickle",
        "_pickle",
        "dill",
        "cloudpickle",
        "joblib",
        # File system operations
        "shutil",
        "tempfile",
        "filecmp",
        "distutils",
        # Shell/terminal
        "pydoc",  # Can run code via pydoc.pager
        "pexpect",
        # Virtual environments (can install packages)
        "venv",
        "ensurepip",
        "pip",
        # Documentation testing (can run code)
        "doctest",
        # IDLE modules (code execution)
        "idlelib",
        # Parser generators (code execution)
        "lib2to3",
        # Network services (constructors bind/listen)
        "socketserver",
        "xmlrpc",
        # Process/thread control
        "signal",
        "_signal",
        "threading",
        "_thread",
        "atexit",
        # Database/file creation
        "sqlite3",
        "_sqlite3",
        # File reading/enumeration
        "fileinput",
        "glob",
        "platform",
        "linecache",
        "difflib",
        # Code compilation (writes .pyc files)
        "compileall",
        "py_compile",
        # Memory mapping
        "mmap",
        # I/O multiplexing (enables network operations)
        "select",
        "selectors",
        # Logging (can open files and network sockets via handlers)
        "logging",
        "syslog",
        # Archive manipulation (can create/extract files)
        "tarfile",
        "zipfile",
        # Shelve (opens database files)
        "shelve",
        # Stdlib modules with hidden subprocess/os.system calls
        "uuid",
        "_osx_support",
        "_aix_support",
    ]
)

BUILTIN_MODULE_NAMES: frozenset[str] = frozenset(["builtins", "__builtins__", "__builtin__"])

# Builtins that are safe to import - pure functions and type constructors
# that cannot be used for code execution or system access.
# Dangerous builtins NOT in this list (and thus blocked):
# - eval, exec, compile: direct code execution
# - open: file system access
# - __import__, __loader__, __spec__: import machinery
# - getattr, setattr, delattr, hasattr: attribute access (can call any method)
# - globals, locals, vars: namespace access
# - input: user input (could be abused)
# - breakpoint: debugger access
# - memoryview: low-level memory access
SAFE_BUILTINS: frozenset[str] = frozenset(
    [
        # Type constructors (create data, cannot execute code)
        "bool",
        "int",
        "float",
        "complex",
        "str",
        "bytes",
        "bytearray",
        "list",
        "tuple",
        "set",
        "frozenset",
        "dict",
        # Pure functions (no side effects, no code execution)
        "len",
        "abs",
        "sum",
        "min",
        "max",
        "round",
        "pow",
        "divmod",
        "sorted",
        "reversed",
        "enumerate",
        "zip",
        "range",
        "map",
        "filter",
        "slice",
        "iter",
        "next",
        "all",
        "any",
        "hash",
        "id",
        "repr",
        "ascii",
        "bin",
        "hex",
        "oct",
        "ord",
        "chr",
        "isinstance",
        "issubclass",
        "object",
        "callable",
        "format",
    ]
)


def is_std_module(module_name: str) -> bool:
    return module_name in BUILTIN_STDLIB_MODULE_NAMES


def extract_identifier_from_ast_node(
    node: ast.expr | str, fallback_prefix: str = "_malformed"
) -> str:
    """
    Extract a valid Python identifier from an AST node.

    For malformed pickle files where STACK_GLOBAL receives complex AST nodes
    instead of strings, this function attempts to extract a meaningful identifier
    or generates a safe fallback.

    Args:
        node: An AST expression node or string
        fallback_prefix: Prefix for generated fallback identifiers

    Returns:
        A valid Python identifier string
    """
    # If already a string, validate and return or fix it
    if isinstance(node, str):
        if node.isidentifier() and not keyword.iskeyword(node):
            return node
        # For invalid string identifiers, create a safe fallback
        node_hash = abs(hash(node)) % 100000
        return f"{fallback_prefix}_str_{node_hash}"

    # Handle ast.Name - extract the id attribute
    if isinstance(node, ast.Name):
        return node.id

    # Handle ast.Attribute - return the attribute name
    if isinstance(node, ast.Attribute):
        return node.attr

    # Handle ast.Call - extract identifier from the function being called
    if isinstance(node, ast.Call):
        return extract_identifier_from_ast_node(node.func, fallback_prefix)

    # Handle ast.Constant - extract value if it's a valid identifier string
    if isinstance(node, ast.Constant):
        value = node.value
        if isinstance(value, str):
            if value.isidentifier() and not keyword.iskeyword(value):
                return value
            # For invalid constant strings, create a deterministic fallback
            value_hash = abs(hash(value)) % 100000
            return f"{fallback_prefix}_const_{value_hash}"
        # For non-string constants (int, float, etc.), create a type-based fallback
        type_name = type(value).__name__
        value_hash = abs(hash(value)) % 100000
        return f"{fallback_prefix}_{type_name}_{value_hash}"

    # For any other complex node types, generate a safe placeholder
    node_type = type(node).__name__.lower()
    node_hash = abs(hash(id(node))) % 100000
    return f"{fallback_prefix}_{node_type}_{node_hash}"


class MarkObject:
    pass


class Opcode:
    name: str
    info: OpcodeInfo

    def __init__(
        self,
        argument: Any | None = None,
        position: int | None = None,
        data: bytes | None = None,
        *,
        info: OpcodeInfo | None = None,
    ):
        if self.__class__ is Opcode:
            if info is None:
                raise TypeError("The Opcode class must be constructed with the `info` argument")
        elif info is not None and info != self.info:
            raise ValueError(
                f"Invalid info type for {self.__class__.__name__}; expected {self.info!r} but got "
                f"{info!r}"
            )
        self.arg: Any = argument
        self.pos: int | None = position
        self._data: bytes | None = data

    def has_data(self) -> bool:
        return self._data is not None

    @property
    def data(self) -> bytes:
        if self._data is None:
            return self.encode()
        return self._data

    @data.setter
    def data(self, value: bytes):
        self._data = value

    def encode(self) -> bytes:
        return self.encode_opcode() + self.encode_body()

    def encode_opcode(self) -> bytes:
        return self.info.code.encode("latin-1")

    def encode_body(self) -> bytes:
        if self.info.arg is None or self.info.arg.n == 0:
            return b""
        raise NotImplementedError(
            f"encode_body() is not yet implemented for opcode {self.__class__.__name__}"
        )

    def __new__(cls, *args, **kwargs):
        if cls is Opcode:
            if "info" not in kwargs:
                raise ValueError(f"You must provide an `info` argument to construct {cls.__name__}")
            info = kwargs["info"]
            if info.name in OPCODES_BY_NAME:
                del kwargs["info"]
                return OPCODES_BY_NAME[info.name](*args, **kwargs)
            raise NotImplementedError(f"TODO: Add support for Opcode {info.name}")
        return super().__new__(cls)

    def run(self, interpreter: Interpreter):
        raise NotImplementedError(f"TODO: Add support for Pickle opcode {self.info.name}")

    def __init_subclass__(cls, **kwargs):
        if cls.__name__ not in (
            "NoOp",
            "StackSliceOpcode",
            "ConstantOpcode",
            "DynamicLength",
            "ConstantInt",
        ):
            if not hasattr(cls, "name") or cls.name is None:
                raise TypeError("Opcode subclasses must define a name")
            if cls.name in OPCODES_BY_NAME:
                raise TypeError(f"An Opcode named {cls.name} is already defined")
            if cls.name not in OPCODE_INFO_BY_NAME:
                raise TypeError(f"An Opcode named {cls.name} is not defined in `pickletools`")
            OPCODES_BY_NAME[cls.name] = cls
            cls.info = OPCODE_INFO_BY_NAME[cls.name]
            # find the associated `pickletools` OpcodeInfo:
        return super().__init_subclass__(**kwargs)

    def __repr__(self):
        if self.pos is None:
            p = ""
        else:
            p = f", position={self.pos!r}"
        if self.has_data():
            d = f", data={self.data!r}"
        else:
            d = ""
        return f"{self.__class__.__name__}(info={self.info!r}, argument={self.arg!r}{d}{p})"


class Endianness(Enum):
    Little = "<"
    Big = ">"


class DynamicLength(Opcode, ABC):
    length_signed: bool = False
    length_bytes: int = 4
    length_endianness: Endianness = Endianness.Little
    struct_types = {1: "b", 2: "h", 4: "i", 8: "q"}
    min_value: int
    max_value: int

    def __init_subclass__(cls, **kwargs):
        ret = super().__init_subclass__(**kwargs)
        length_bits = cls.length_bytes * 8
        if cls.length_signed:
            cls.min_value = 1 << (length_bits - 1)
            cls.max_value = cls.min_value ^ (2**length_bits - 1)
        else:
            cls.min_value = 0
            cls.max_value = 2**length_bits - 1
        return ret

    @classmethod
    def encode_length(cls, length: int) -> bytes:
        if cls.length_bytes not in cls.struct_types:
            raise TypeError(
                f"{cls.__name__}.struct_types does not include a value for "
                f"{cls.__name__}.length_bytes = {cls.length_bytes}"
            )
        if length < cls.min_value or length > cls.max_value:
            raise ValueError(
                f"Invalid length {length}: {cls.__name__} can only represent lengths in the range "
                f"[{cls.min_value}, {cls.max_value}]"
            )
        st = cls.struct_types[cls.length_bytes]
        if not cls.length_signed:
            st = st.upper()
        return struct.pack(f"{cls.length_endianness.value}{st}", length)

    def encode(self) -> bytes:
        body = self.encode_body()
        return self.encode_opcode() + self.encode_length(len(body)) + body

    @classmethod
    def validate(cls, obj):
        length = len(cls(obj).encode_body())
        if length < cls.min_value or length > cls.max_value:
            raise ValueError(
                f"Invalid object {obj!r}: {cls.__name__} can only represent objects with lengths "
                f"in the range [{cls.min_value}, {cls.max_value}]"
            )
        return obj


class NoOp(Opcode):
    def run(self, interpreter: Interpreter):
        pass


def raw_unicode_escape(byte_string: bytes) -> str:
    s = []
    for b in byte_string:
        if 32 <= b <= 128:
            # this is printable ASCII
            s.append(chr(b))
        elif b == ord("\n"):
            s.append("\\n")
        elif b == ord("\r"):
            s.append("\\r")
        elif b == ord("\\"):
            s.append("\\\\")
        else:
            s.append(f"\\u{b:04x}")
    s.append("\n")
    return "".join(s)


class ConstantOpcode(Opcode):
    ConstantOpcodePriorities: dict[type[ConstantOpcode], int] = {}
    priority: int

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(make_constant(self.arg))

    def __init_subclass__(cls, **kwargs):
        ret = super().__init_subclass__(**kwargs)
        if not cls.__name__ == "ConstantInt":
            if cls.validate.__code__ == ConstantOpcode.validate.__code__:
                raise TypeError(f"{cls.__name__} must implement the validate method")
            if (
                not hasattr(cls, "priority")
                or not isinstance(cls.priority, int)
                or cls.priority is None
            ):
                raise TypeError(
                    f"{cls.__name__} must define an integer priority used for auto-instantiation "
                    "from ConstantOpcode.new(...)"
                )
            ConstantOpcode.ConstantOpcodePriorities[cls] = cls.priority
        return ret

    @classmethod
    def validate(cls, obj):
        """
        Validates whether obj can be used to instantiate a new instance of this class using new(...)

        Returning the value of the object to be saved to the constant
        Or throwing a ValueError if obj cannot be used to instantiate this type of constant
        """
        raise NotImplementedError()

    @classmethod
    def new(cls: type[T], obj) -> T:
        for subclass, _ in sorted(
            ConstantOpcode.ConstantOpcodePriorities.items(), key=lambda kv: kv[1]
        ):
            if not issubclass(subclass, cls):
                continue
            try:
                return subclass(subclass.validate(obj))
            except ValueError:
                pass
        raise ValueError(
            "There is no subclass of ConstantOpcode that handles objects of type "
            f"{type(obj)!r} for {obj!r}"
        )


class ConstantInt(ConstantOpcode, ABC):
    signed: bool = False
    num_bytes: int = 4
    endianness: Endianness = Endianness.Little
    struct_types = {1: "b", 2: "h", 4: "i", 8: "q"}
    min_value: int
    max_value: int

    def __init_subclass__(cls, **kwargs):
        ret = super().__init_subclass__(**kwargs)
        length_bits = cls.num_bytes * 8
        if cls.signed:
            cls.min_value = 1 << (length_bits - 1)
            cls.max_value = cls.min_value ^ (2**length_bits - 1)
        else:
            cls.min_value = 0
            cls.max_value = 2**length_bits - 1
        return ret

    def encode_body(self) -> bytes:
        st = self.struct_types[self.num_bytes]
        if not self.signed:
            st = st.upper()
        return struct.pack(f"{self.endianness.value}{st}", self.arg)

    @classmethod
    def validate(cls, obj):
        if not isinstance(obj, int):
            raise ValueError(f"{cls.__name__} can only be instantiated from integers, not {obj!r}")
        if cls.num_bytes not in cls.struct_types:
            raise TypeError(
                f"{cls.__name__}.struct_types does not include a value for "
                f"{cls.__name__}.length_bytes = {cls.num_bytes}"
            )
        if obj < cls.min_value or obj > cls.max_value:
            raise ValueError(
                f"Invalid value {obj!r}: {cls.__name__} can only represent lengths in the range "
                f"[{cls.min_value}, {cls.max_value}]"
            )
        return obj


class StackSliceOpcode(Opcode):
    def run(self, interpreter: Interpreter, stack_slice: list[ast.expr]):
        raise NotImplementedError(f"{self.__class__.__name__} must implement run()")

    def __init_subclass__(cls, **kwargs):
        ret = super().__init_subclass__(**kwargs)
        orig_run = cls.run

        def run_wrapper(self, interpreter: Interpreter):
            args = []
            while True:
                if not interpreter.stack:
                    raise InterpretationError(
                        "Exhausted the stack while searching for a MarkObject!"
                    )
                obj = interpreter.stack.pop()
                if isinstance(obj, MarkObject):
                    break
                args.append(obj)
            args = list(reversed(args))
            return orig_run(self, interpreter, args)

        cls.run = run_wrapper

        return ret


class ASTProperties(ast.NodeVisitor):
    def __init__(self):
        self.imports: list[ast.Import | ast.ImportFrom] = []
        self.calls: list[ast.Call] = []
        self.non_setstate_calls: list[ast.Call] = []
        self.likely_safe_imports: set[str] = set()
        self._visited: set[int] = set()  # Track visited nodes by id to detect cycles

    def visit(self, node: ast.AST) -> Any:
        """Override visit to detect and skip cycles in the AST.

        Pickle files can create cyclic AST structures via MEMOIZE + GET opcodes.
        Without cycle detection, visiting such structures causes infinite recursion.
        """
        node_id = id(node)
        if node_id in self._visited:
            return None  # Skip already-visited nodes
        self._visited.add(node_id)
        return super().visit(node)

    def _process_import(self, node: ast.Import | ast.ImportFrom):
        self.imports.append(node)
        if (
            isinstance(node, ast.ImportFrom)
            and node.module is not None
            and is_std_module(node.module)
        ):
            self.likely_safe_imports |= {name.name for name in node.names}

    def visit_Import(self, node: ast.Import):
        self._process_import(node)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        self._process_import(node)

    def visit_Call(self, node: ast.Call):
        self.calls.append(node)
        if not isinstance(node.func, ast.Attribute) or node.func.attr != "__setstate__":
            self.non_setstate_calls.append(node)


class PickleDecodeError(ValueError):
    pass


class EmptyPickleError(PickleDecodeError):
    pass


class InterpretationError(PickleDecodeError):
    """Raised when pickle interpretation fails due to malformed opcode sequences."""

    pass


class Pickled(OpcodeSequence):
    def __init__(self, opcodes: Iterable[Opcode], has_invalid_opcode: bool = False):
        self._opcodes: list[Opcode] = list(opcodes)
        self._ast: ast.Module | None = None
        self._properties: ASTProperties | None = None
        # Whether the pickled sequence was interrupted because of
        # an invalid opcode
        self._has_invalid_opcode: bool = has_invalid_opcode
        # Whether the pickle contains cyclic references
        self._has_cycles: bool = False
        self._has_interpretation_error: bool = False
        self._has_resource_exhaustion: bool = False

    def __len__(self) -> int:
        return len(self._opcodes)

    def __iter__(self) -> Iterator[Opcode]:
        return iter(self._opcodes)

    def __getitem__(self, index: int) -> Opcode:
        return self._opcodes[index]

    def insert(self, index: int, opcode: Opcode):
        self._opcodes.insert(index, opcode)
        self._ast = None
        self._properties = None

    def _is_constant_type(self, obj: Any) -> bool:
        return isinstance(obj, int | float | str | bytes)

    def _encode_python_obj(self, obj: Any) -> List[Opcode]:
        """Create an opcode sequence that builds an arbitrary python object on the top of the
        pickle VM stack"""
        if self._is_constant_type(obj):
            return [ConstantOpcode.new(obj)]
        if isinstance(obj, list):
            res = [Mark()]
            for item in obj:
                if self._is_constant_type(item):
                    res.append(ConstantOpcode.new(item))
                else:
                    res += self._encode_python_obj(item)
            res.append(List())
            return res
        if isinstance(obj, dict):
            if len(obj) == 0:
                res = [EmptyDict()]
            else:
                res = [Mark()]
                for key, val in obj.items():
                    res.append(ConstantOpcode.new(key))  # Assume key is constant
                    if self._is_constant_type(val):
                        res.append(ConstantOpcode.new(val))
                    else:
                        res += self._encode_python_obj(val)
                res.append(Dict())
            return res
        raise ValueError(f"Type {type(obj)} not supported")

    def insert_python_obj(self, index: int, obj: Any) -> int:
        """Insert an opcode sequence that constructs a python object on the stack.
        Returns the number of opcodes inserted"""
        opcodes = self._encode_python_obj(obj)
        for i, opcode in enumerate(opcodes):
            self.insert(index + i, opcode)
        return len(opcodes)

    def insert_python(
        self,
        *args,
        module: str = "builtins",
        attr: str = "eval",
        run_first: bool = True,
        use_output_as_unpickle_result: bool = False,
    ) -> int:
        if not isinstance(self[-1], Stop):
            raise ValueError("Expected the last opcode to be STOP")
        # we need to add the call to GLOBAL before the preexisting code, because the following code
        # can sometimes mess up module lookup (somehow? I, Evan, don't fully understand why yet).
        # So we set up the "import" of `__builtin__.eval` first, then set up the stack for a call
        # to it, and then either immediately call the `eval` with a `Reduce` opcode (the default)
        # or optionally insert the `Reduce` at the end (and hope that the existing code cleans up
        # its stack so it remains how we left it!
        # TODO: Add code to emulate the code afterward and confirm that the stack is sane!
        i = 0
        while isinstance(self[i], Proto | Frame):
            i += 1
        self.insert(i, Global.create(module, attr))
        i += 1
        self.insert(i, Mark())
        i += 1
        for arg in args:
            i += self.insert_python_obj(i, arg)
            # self.insert(i, ConstantOpcode.new(arg))
            # i += 1
        self.insert(i, Tuple())
        i += 1
        if run_first:
            self.insert(i, Reduce())
            if use_output_as_unpickle_result:
                self.insert(-1, Pop())  # Just discard the original unpickle
            else:
                # At the end, the stack will contain [reduce_res, original_obj].
                # We need to remove everything below original_obj:
                # NOTE(boyan): using an arbitrary MEMO key here seems to work. If not,
                # then switch to using the interpreter() to determine the correct MEMO key to use here
                self.insert(-1, Put(321987))  # Put obj in memo
                self.insert(-1, Pop())  # Pop obj and reduce_res under
                self.insert(-1, Pop())
                self.insert(-1, Get.create(321987))  # Get back obj
            return i + 1
        # Inject call
        if use_output_as_unpickle_result:
            # the top of the stack should be the original unpickled value, but we can throw
            # that away because we are replacing it with the result of calling eval:
            self.insert(-1, Pop())
            # now the top of the stack should be our original Global, Mark, Unicode,
            # Tuple setup, ready for Reduce:
            self.insert(-1, Reduce())
        else:
            # we need to preserve the "real" output of the preexisting unpickling, which should
            # be at the top of the stack, directly above our Tuple, Unicode, Mark, and Global
            # stack items we added above.
            # So, we have to save the original result to the memo. First, interpret the existing
            # code to see which memo location it would be saved to:
            interpreter = Interpreter(self)
            interpreter.run()
            memo_id = len(interpreter.memory)
            self.insert(-1, Memoize())
            self.insert(-1, Pop())
            self.insert(-1, Reduce())
            self.insert(-1, Pop())
            self.insert(-1, Get.create(memo_id))
        return -1

    insert_python_eval = insert_python

    def append_python(
        self,
        *args,
        module: str = "builtins",
        attr: str = "eval",
        pop_result: bool = False,
    ) -> None:
        """Append python code to run at the end of the pickle.

        :param pop_result: whether to pop the result of the run off the stack. Appends
        a POP instruction if True"""
        if not isinstance(self[-1], Stop):
            raise ValueError("Expected the last opcode to be STOP")
        # NOTE(boyan): this seems to work even without insert GLOBAL at the beginning
        # of the pickle, but see comment in 'insert_python'
        self.insert(-1, Global.create(module, attr))
        self.insert(-1, Mark())
        for arg in args:
            self.insert(-1, ConstantOpcode.new(arg))
        self.insert(-1, Tuple())
        self.insert(-1, Reduce())
        if pop_result:
            self.insert(-1, Pop())

    def insert_magic_int(self, magic: int, index: int = -1):
        """Insert and pop a specific integer value. This is used for persistent
        injections to locate the injection payload in the pickled file. The value
        is artificially added by using a dummy INT + POP combination that doesn't
        affect the stack when executed

        :param magic: magic integer value to add
        :param index: index in opcodes list where to insert the magic"""
        self.insert(index, Int(magic))
        self.insert(-1 if index == -1 else index + 1, Pop())

    def insert_function_call_on_unpickled_object(
        self,
        function_definition: str,
        constant_args: List[Any] | None = None,
        compile_code: bool = False,
    ):
        """Insert and call a function that takes the unpickled object as parameter.

        :param function_definition: a string containing the full python definiton of the function
        to call, including the `def` keyword. The function prototype must be `myfunc(obj)` where
        `obj` is the object being unpickled. The function return value is used as the unpickling
        output.

        :param compile_code: whether the function definition should be precompiled into
        Python bytecode, for increased obfuscation. Note that the function name will still be
        exposed as plaintext sourcecode as this is required to make the function callable through
        a call to (pseudocode) "eval(function_name)".
        """

        if not isinstance(self[-1], Stop):
            raise ValueError("Expected the last opcode to be STOP")

        # Get function name
        fn_match = list(re.match(r"def\s+(.*?)\s*\(", function_definition).groups())
        if not fn_match:
            raise ValueError("Failed to extract function name from function definition")
        function_name = fn_match[0]

        # Insert exec of the function definition in advance
        if compile_code:
            # Compile the function definition
            code_obj = compile(function_definition, "<string>", "exec")
            bytecode = marshal.dumps(code_obj)  # marshal is required to get literal bytes

            # Instructions:
            ## Current stack status: [model]
            ## Add the compiled bytes to the stack, then unmarshal them back into a code object
            self.append_python(bytecode, module="marshal", attr="loads", pop_result=False)
            ## [model, def func]
            ## Slide exec instructions under the function definition on the stack
            self.insert(-1, Put(1))  # Move function def off the stack and into memory
            self.insert(-1, Pop())
            ## [model]
            if not isinstance(self[-1], Stop):  # sanity check akin to the one in append_python
                raise ValueError("Expected the last opcode to be STOP")
            ### NOTE(boyan): this seems to work even without insert GLOBAL at the beginning
            ### of the pickle, but see comment in 'insert_python'
            self.insert(-1, Global.create("builtins", "exec"))
            ## [model, exec]
            self.insert(-1, Mark())
            ## [model, exec, mark]
            self.insert(-1, Get.create(1))  # Place the function def back
            ## [model, exec, mark, def func]
            self.insert(-1, Tuple())
            ## [model, exec, (def func)]
            self.insert(-1, Reduce())
            ## [model, exec return value]
            self.insert(-1, Pop())  # Remove extraneous return value from the stack
            ## [model]
        else:
            ## Current stack status: [model]
            self.append_python(function_definition, attr="exec", pop_result=True)
            ## [model]

        # Eval the function name get the callable object
        # If we inject myfunc() this will return eval(myfunc) which is the myfunc callable object
        self.append_python(function_name, attr="eval")
        # [model, func]

        # At the end of all of the above operations, the stack contains [model, func].
        # We swap them on the stack:
        self.insert(-1, Put(1))  # Put func in memo
        self.insert(-1, Pop())
        # [model]
        self.insert(-1, Put(2))  # Put model in memo
        self.insert(-1, Pop())
        # []
        self.insert(-1, Get.create(1))
        # [func]
        self.insert(-1, Mark())
        # [func, mark]
        self.insert(-1, Get.create(2))
        # [func, mark, model]

        # Add constant arguments
        if constant_args:
            for arg in constant_args:
                self.insert(-1, ConstantOpcode.new(arg))
        # [func, mark, model, arg1, ..., argn]

        # We need to add TUPLE which
        # packs the function arguments from the stack and then call REDUCE, which calls the injected
        # function.
        # Note: precondition says the function must return the final object so no need to save the
        # object before calling reduce.
        self.insert(-1, Tuple())
        # [func, (model, arg1, ..., argn)]
        self.insert(-1, Reduce())
        # [func return value]

    def insert_python_exec(
        self,
        *args,
        run_first: bool = True,
        use_output_as_unpickle_result: bool = False,
    ):
        return self.insert_python(
            *args,
            module="builtins",
            attr="exec",
            run_first=run_first,
            use_output_as_unpickle_result=use_output_as_unpickle_result,
        )

    def __setitem__(self, index: int | slice, item: Opcode | Iterable[Opcode]):
        self._opcodes[index] = item
        self._ast = None
        self._properties = None

    def __delitem__(self, index: int):
        del self._opcodes[index]
        self._ast = None
        self._properties = None

    def dumps(self) -> bytes:
        b = bytearray()
        for opcode in self:
            b.extend(opcode.data)
        return bytes(b)

    def dump(self, file: BinaryIO):
        for opcode in self:
            file.write(opcode.data)

    def dumps_partial(self, from_idx: int, to_idx: int) -> bytes:
        """Dump bytecode only between two opcodes

        :param from_idx: index of opcode from which we dump (included)
        :param to_idx: index of opcode until which we dump (included). If -1, dump to end of file
        """
        # Sanity check
        assert from_idx >= 0 and (to_idx >= from_idx or to_idx == -1)

        b = bytearray()
        for opcode in self._opcodes[from_idx:to_idx]:
            b.extend(opcode.data)
        return bytes(b)

    @property
    def opcodes(self) -> Iterator[Opcode]:
        return iter(self)

    @property
    def has_invalid_opcode(self) -> bool:
        return self._has_invalid_opcode

    @property
    def has_interpretation_error(self) -> bool:
        _ = self.ast  # Ensure interpretation ran
        return self._has_interpretation_error

    @property
    def has_resource_exhaustion(self) -> bool:
        _ = self.ast  # Ensure interpretation ran
        return self._has_resource_exhaustion

    @staticmethod
    def make_stream(data: Buffer | BinaryIO) -> BinaryIO:
        if isinstance(data, bytes | bytearray | Buffer):
            data = BytesIO(data)
        elif (not hasattr(data, "seekable") or not data.seekable()) and hasattr(data, "read"):
            data = BytesIO(data.read())
        return data

    @staticmethod
    def load(pickled: Buffer | BinaryIO, fail_on_decode_error: bool = True) -> Pickled:
        """
        :arg fail_on_decode_error: If a decoding error such as unknown opcode happens
            and this parameter is set to False, then the error is ignored and the
            returns the list of opcodes decoded until the error. If set to True,
            then raises a decode error exception
        """

        pickled = Pickled.make_stream(pickled)
        first_pos = pickled.tell()
        opcodes: List[Opcode] = []
        has_invalid_opcode = False

        try:
            for info, arg, pos in genops(pickled):
                pos_before = pickled.tell()
                try:
                    if (
                        pos is not None
                        and opcodes
                        and opcodes[-1].pos is not None
                        and not opcodes[-1].has_data()
                        and opcodes[-1].pos < pos
                    ):
                        pickled.seek(opcodes[-1].pos)
                        opcodes[-1].data = pickled.read(pos - opcodes[-1].pos)
                    if pos is not None:
                        pickled.seek(pos)
                    if info.arg is None or info.arg.n == 0:
                        if pos is not None:
                            data = None
                        else:
                            data = info.code.encode("utf-8")
                    elif info.arg.n > 0 and pos is not None:
                        data = pickled.read(len(info.code) + info.arg.n)
                        if len(data) != len(info.code) + info.arg.n:
                            raise PickleDecodeError(
                                f"Error decoding opcode {info.name} at offset {pos}: "
                                f"Expected {len(info.code) + info.arg.n} bytes of data but only "
                                f"read {len(data)}"
                            )
                    else:
                        data = None
                    opcodes.append(Opcode(info=info, argument=arg, data=data, position=pos))
                finally:
                    # Need to reset the position within the file so as not to confuse genops
                    pickled.seek(pos_before)
        except ValueError as e:
            if opcodes:
                if fail_on_decode_error:
                    raise PickleDecodeError(e)
                has_invalid_opcode = True
            else:
                raise EmptyPickleError()
        if opcodes:
            if opcodes[-1].pos is not None:
                if opcodes[-1].has_data():
                    last_pos = opcodes[-1].pos + len(opcodes[-1].data)
                else:
                    last_pos = opcodes[-1].pos + len(opcodes[-1].info.code)
                pickled.seek(last_pos)
        else:
            pickled.seek(first_pos)
        return Pickled(opcodes, has_invalid_opcode=has_invalid_opcode)

    @property
    def properties(self) -> ASTProperties:
        if self._properties is None:
            self._properties = ASTProperties()
            self._properties.visit(self.ast)
        return self._properties

    @property
    def has_import(self) -> bool:
        """Checks whether unpickling would cause an import to be run"""
        return bool(self.properties.imports)

    @property
    def has_call(self) -> bool:
        """Checks whether unpickling would cause a function call"""
        return bool(self.properties.calls)

    @property
    def has_non_setstate_call(self) -> bool:
        """Checks whether unpickling would cause a call to a function other than
        object.__setstate__"""
        return bool(self.properties.non_setstate_calls)

    def check_safety(self):
        raise WrongMethodError(
            """This method has been removed. Use fickling.analysis.check_safety()
on the Pickled object instead"""
        )

    def is_likely_safe(self):
        raise WrongMethodError(
            "This method has been removed. Use fickling.is_likely_safe() on the pickle file instead"
        )

    def unsafe_imports(self) -> Iterator[ast.Import | ast.ImportFrom]:
        for node in self.properties.imports:
            if isinstance(node, ast.ImportFrom):
                if node.module and any(
                    component in UNSAFE_IMPORTS for component in node.module.split(".")
                ):
                    yield node
                elif "eval" in (n.name for n in node.names):
                    yield node
            else:
                # ast.Import: check if any imported module is unsafe
                if any(
                    any(component in UNSAFE_IMPORTS for component in alias.name.split("."))
                    for alias in node.names
                ):
                    yield node
                elif "eval" in (alias.name for alias in node.names):
                    yield node

    def non_standard_imports(self) -> Iterator[ast.Import | ast.ImportFrom]:
        for node in self.properties.imports:
            if isinstance(node, ast.ImportFrom):
                # from module import x - check if module is standard
                # module can be None for relative imports (from . import x)
                if node.module is None or not is_std_module(node.module):
                    yield node
            else:
                # import x, y, z - check if any name is non-standard
                if any(not is_std_module(alias.name) for alias in node.names):
                    yield node

    @property
    def ast(self) -> ast.Module:
        if self._ast is None:
            try:
                interpreter = Interpreter(self)
                self._ast = interpreter.to_ast()
                self._has_cycles = interpreter._has_cycle
            except InterpretationError as e:
                self._has_interpretation_error = True
                sys.stderr.write(
                    f"Warning: malformed pickle file. {e!s}; "
                    f"returning empty AST to continue analysis\n"
                )
                self._ast = ast.Module(body=[], type_ignores=[])
            except ResourceExhaustionError:
                self._has_resource_exhaustion = True
                sys.stderr.write(
                    "Warning: resource limits exceeded during interpretation; "
                    "returning empty AST to continue analysis\n"
                )
                self._ast = ast.Module(body=[], type_ignores=[])
        return self._ast

    @property
    def has_cycles(self) -> bool:
        """Check if the pickle contains cyclic references."""
        _ = self.ast  # Ensure interpretation ran
        return self._has_cycles

    @property
    def nb_opcodes(self) -> int:
        return len(self._opcodes)


class Stack(GenericSequence, Generic[T]):
    def __init__(self, initial_value: Iterable[T] = ()):
        self._stack: List[T] = list(initial_value)
        self.opcode: Opcode | None = None

    @overload
    @abstractmethod
    def __getitem__(self, i: int) -> T: ...

    @overload
    @abstractmethod
    def __getitem__(self, s: slice) -> GenericSequence: ...

    def __getitem__(self, i: int) -> T:
        return self._stack[i]

    def __len__(self) -> int:
        return len(self._stack)

    def pop(self):
        if not self._stack:
            if self.opcode is None:
                raise IndexError("Stack is empty")
            raise IndexError(f"Opcode {self.opcode!s} attempted to pop from an empty stack")
        return self._stack.pop()

    def push(self, obj: T):
        self._stack.append(obj)

    append = push

    def __str__(self):
        return str(self._stack)

    def __repr__(self):
        return f"{self.__class__.__name__}(initial_value={self._stack!r})"


class ModuleBody:
    def __init__(self, interpreter: Interpreter):
        self._list: List[ast.stmt] = []
        self.interpreter: Interpreter = interpreter

    def append(self, stmt: ast.stmt):
        lineno = len(self._list) + 1
        if hasattr(stmt, "lineno") and stmt.lineno is not None and stmt.lineno != lineno:
            raise ValueError(
                f"Statement {stmt} was expected to have line number {lineno} but instead has "
                f"{stmt.lineno}"
            )
        stmt.lineno = lineno
        self._list.append(stmt)

    def extend(self, stmts: Iterable[ast.stmt]):
        for stmt in stmts:
            self.append(stmt)

    def __iter__(self) -> Iterator[ast.stmt]:
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def __getitem__(self, index: int | slice) -> ast.stmt:
        return self._list[index]


class Interpreter:
    def __init__(
        self,
        pickled: Pickled,
        first_variable_id: int = 0,
        result_variable: str = "result",
        limits: InterpreterLimits | None = None,
    ):
        self.pickled: Pickled = pickled
        self.memory: dict[int, ast.expr] = {}
        self.stack: Stack[ast.expr | MarkObject] = Stack()
        self.module_body: ModuleBody = ModuleBody(self)
        self.result_variable: str = result_variable
        self._module: ast.Module | None = None
        self._var_counter: int = first_variable_id
        self._opcodes: Iterator[Opcode] = iter(pickled)
        self._has_cycle: bool = False

        # Resource limits and tracking for DoS protection
        self.limits: InterpreterLimits = limits or DEFAULT_INTERPRETER_LIMITS
        self._opcode_count: int = 0
        self._get_count: int = 0
        self._put_count: int = 0

    @property
    def next_variable_id(self) -> int:
        return self._var_counter

    def to_ast(self) -> ast.Module:
        if self._module is None:
            self.run()
        return self._module

    def unused_assignments(self) -> dict[str, ast.Assign]:
        if self._module is None:
            self.run()
        used: set[str] = set()
        defined: set[str] = set()
        assignments: dict[str, ast.Assign] = {}
        for statement in self.module_body:
            # skip the last statement because it is always used
            if isinstance(statement, ast.Assign):
                if (
                    len(statement.targets) == 1
                    and isinstance(statement.targets[0], ast.Name)
                    and statement.targets[0].id == "result"
                ):
                    # this is the return value of the program
                    break
                for target in statement.targets:
                    if isinstance(target, ast.Name):
                        defined.add(target.id)
                        if target.id in assignments:
                            # this should never happen, since Fickling constructs the AST
                            sys.stderr.write(
                                f"Warning: Duplicate declaration of variable {target.id}\n"
                            )
                        assignments[target.id] = statement
                statement = statement.value
            if statement is not None:
                for node in ast.walk(statement):
                    if isinstance(node, ast.Name):
                        used.add(node.id)
        return {varname: assignments[varname] for varname in defined - used}

    def unused_variables(self) -> frozenset[str]:
        return self.unused_assignments().keys()  # type: ignore

    def stop(self):
        self._opcodes = iter(())

    def run(self):
        while True:
            try:
                self.step()
            except StopIteration:
                break

    def step(self) -> Opcode:
        try:
            opcode = next(self._opcodes)
            finished = False
        except StopIteration:
            # we finished running the program
            finished = True
        if finished:
            for i, stmt in enumerate(self.module_body):
                stmt.lineno = i + 1
                stmt.col_offset = 0
            self._module = ast.Module(list(self.module_body), type_ignores=[])
            raise StopIteration()
        self._opcode_count += 1
        self.stack.opcode = opcode
        opcode.run(self)
        self._check_limits()
        return opcode

    def _check_limits(self) -> None:
        """Check resource limits and raise if exceeded."""
        if self._opcode_count > self.limits.max_opcodes:
            raise ResourceExhaustionError("opcodes", self.limits.max_opcodes, self._opcode_count)
        if len(self.stack) > self.limits.max_stack_depth:
            raise ResourceExhaustionError(
                "stack_depth", self.limits.max_stack_depth, len(self.stack)
            )
        if len(self.memory) > self.limits.max_memo_size:
            raise ResourceExhaustionError("memo_size", self.limits.max_memo_size, len(self.memory))
        # Check for suspicious GET/PUT ratio (expansion attack indicator)
        if self._put_count > 0:
            ratio = self._get_count / self._put_count
            if ratio > self.limits.max_get_ratio:
                raise ExpansionAttackError(self.limits.max_get_ratio, round(ratio))

    def track_get(self) -> None:
        """Track a GET operation for DoS protection."""
        self._get_count += 1

    def track_put(self) -> None:
        """Track a PUT operation for DoS protection."""
        self._put_count += 1

    def new_variable(self, value: ast.expr, name: str | None = None) -> str:
        if name is None:
            name = f"_var{self._var_counter}"
            self._var_counter += 1
        self.module_body.append(ast.Assign([ast.Name(name, ast.Store())], value))
        return name

    @staticmethod
    def interpret(pickled: Pickled) -> ast.Module:
        return Interpreter(pickled).to_ast()

    def __str__(self):
        return ast.dump(self.to_ast())


class Proto(NoOp):
    name = "PROTO"

    @staticmethod
    def create(version: int) -> Proto:
        return Proto(version)

    def encode_body(self) -> bytes:
        return bytes([self.version])

    @property
    def version(self) -> int:
        if self.arg is None:
            return 0
        if isinstance(self.arg, int):
            return self.arg
        # Endianness shouldn't really matter here because there is only one byte for the version
        return int.from_bytes(self.arg, "big", signed=False)


class Global(Opcode):
    name = "GLOBAL"

    @staticmethod
    def create(module: str, attr: str) -> Global:
        return Global(f"{module} {attr}")

    @property
    def module(self) -> str:
        return next(iter(self.arg.split(" ")))

    @property
    def attr(self) -> str:
        _, attribute, *_ = self.arg.split(" ")
        return attribute

    def run(self, interpreter: Interpreter):
        module, attr = self.module, self.attr
        alias = ast.alias(attr)
        interpreter.module_body.append(ast.ImportFrom(module=module, names=[alias], level=0))
        interpreter.stack.append(ast.Name(attr, ast.Load()))

    def encode(self) -> bytes:
        return f"c{self.module}\n{self.attr}\n".encode()


class StackGlobal(NoOp):
    name = "STACK_GLOBAL"

    def run(self, interpreter: Interpreter):
        attr = interpreter.stack.pop()
        module = interpreter.stack.pop()

        # Extract values from ast.Constant nodes
        if isinstance(module, ast.Constant):
            module = module.value
        if isinstance(attr, ast.Constant):
            attr = attr.value

        # Normalize module and attr to strings, extracting meaningful identifiers from AST nodes
        module_needs_extraction = not isinstance(module, str)
        attr_needs_extraction = not isinstance(attr, str)

        if module_needs_extraction or attr_needs_extraction:
            sys.stdout.write(
                f"Warning: malformed pickle file. STACK_GLOBAL called with invalid types. "
                f"'Module' is {type(module).__name__} ({module!r}), 'Attr' is {type(attr).__name__} ({attr!r}). "
                f"Expected str; extracting identifiers to continue analysis.\n"
            )

            if module_needs_extraction:
                module = extract_identifier_from_ast_node(
                    module, fallback_prefix="_malformed_module"
                )
            if attr_needs_extraction:
                attr = extract_identifier_from_ast_node(attr, fallback_prefix="_malformed_attr")

        # Final validation: ensure both are valid identifier strings
        if not isinstance(module, str) or not isinstance(attr, str):
            raise TypeError(
                f"Failed to extract valid identifiers from STACK_GLOBAL arguments. "
                f"Module: {type(module).__name__}, Attr: {type(attr).__name__}"
            )

        if not all(m.isidentifier() for m in module.split(".")) or not all(
            a.isidentifier() for a in attr.split(".")
        ):
            raise ValueError(
                f"Extracted identifiers are not valid Python identifiers. "
                f"Module: {module!r}, Attr: {attr!r}"
            )

        alias = ast.alias(attr)
        interpreter.module_body.append(ast.ImportFrom(module=module, names=[alias], level=0))
        interpreter.stack.append(ast.Name(attr, ast.Load()))


class Ext1(Opcode):
    name = "EXT1"

    def run(self, interpreter: Interpreter):
        code = self.arg

        # Generate: import copyreg, sys
        interpreter.module_body.append(
            ast.Import(names=[ast.alias("copyreg", None), ast.alias("sys", None)])
        )

        # Look up (module, name) from _inverted_registry
        # Generate: _reg_entry = copyreg._inverted_registry.get(code, (None, None))
        lookup_call = ast.Call(
            ast.Attribute(
                ast.Attribute(ast.Name("copyreg", ast.Load()), "_inverted_registry", ast.Load()),
                "get",
                ast.Load(),
            ),
            [
                ast.Constant(code),
                ast.Tuple([ast.Constant(None), ast.Constant(None)], ast.Load()),
            ],
            [],
        )
        registry_var = interpreter.new_variable(lookup_call)

        # Generate: __import__(_reg_entry[0])
        import_call = ast.Call(
            ast.Name("__import__", ast.Load()),
            [ast.Subscript(ast.Name(registry_var, ast.Load()), ast.Constant(0), ast.Load())],
            [],
        )
        interpreter.module_body.append(ast.Expr(import_call))

        # Generate: var = getattr(sys.modules[_reg_entry[0]], _reg_entry[1])
        getattr_call = ast.Call(
            ast.Name("getattr", ast.Load()),
            [
                ast.Subscript(
                    ast.Attribute(ast.Name("sys", ast.Load()), "modules", ast.Load()),
                    ast.Subscript(ast.Name(registry_var, ast.Load()), ast.Constant(0), ast.Load()),
                    ast.Load(),
                ),
                ast.Subscript(ast.Name(registry_var, ast.Load()), ast.Constant(1), ast.Load()),
            ],
            [],
        )
        var_name = interpreter.new_variable(getattr_call)
        interpreter.stack.append(ast.Name(var_name, ast.Load()))


class Ext2(Ext1):
    name = "EXT2"


class Ext4(Ext1):
    name = "EXT4"


class Inst(StackSliceOpcode):
    name = "INST"

    @staticmethod
    def create(module: str, classname: str) -> Inst:
        return Inst(f"{module} {classname}")

    @property
    def module(self) -> str:
        return next(iter(self.arg.split(" ")))

    @property
    def cls(self) -> str:
        _, classname, *_ = self.arg.split(" ")
        return classname

    def run(self, interpreter: Interpreter, stack_slice: List[ast.expr]):
        module, classname = self.module, self.cls
        alias = ast.alias(classname)
        interpreter.module_body.append(ast.ImportFrom(module=module, names=[alias], level=0))
        args = ast.Tuple(tuple(stack_slice))
        call = ast.Call(ast.Name(classname, ast.Load()), list(args.elts), [])
        var_name = interpreter.new_variable(call)
        interpreter.stack.append(ast.Name(var_name, ast.Load()))

    def encode(self) -> bytes:
        return f"i{self.module}\n{self.classname}\n".encode()


class Put(Opcode):
    name = "PUT"

    def run(self, interpreter: Interpreter):
        interpreter.track_put()
        interpreter.memory[self.arg] = interpreter.stack[-1]

    def encode_body(self) -> bytes:
        # Encode memo_id as decimal \n terminated string
        return f"{self.arg}\n".encode()


class BinPut(Opcode):
    name = "BINPUT"

    def run(self, interpreter: Interpreter):
        interpreter.track_put()
        interpreter.memory[self.arg] = interpreter.stack[-1]

    def encode_body(self):
        assert self.arg <= 255, "BINPUT only supports 1-byte memo indexing"
        return bytes([self.arg])


class LongBinPut(BinPut):
    name = "LONG_BINPUT"


class EmptyTuple(Opcode):
    name = "EMPTY_TUPLE"

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(ast.Tuple((), ast.Load()))


class TupleOne(Opcode):
    name = "TUPLE1"

    def run(self, interpreter: Interpreter):
        stack_top = interpreter.stack.pop()
        interpreter.stack.push(ast.Tuple((stack_top,), ast.Load()))


class TupleTwo(Opcode):
    name = "TUPLE2"

    def run(self, interpreter: Interpreter):
        arg2 = interpreter.stack.pop()
        arg1 = interpreter.stack.pop()
        interpreter.stack.append(ast.Tuple((arg1, arg2), ast.Load()))


class TupleThree(Opcode):
    name = "TUPLE3"

    def run(self, interpreter: Interpreter):
        top = interpreter.stack.pop()
        mid = interpreter.stack.pop()
        bot = interpreter.stack.pop()
        interpreter.stack.append(ast.Tuple((bot, mid, top), ast.Load()))


class AddItems(Opcode):
    name = "ADDITEMS"

    def run(self, interpreter: Interpreter):
        to_add = []
        while interpreter.stack:
            obj = interpreter.stack.pop()
            if isinstance(obj, MarkObject):
                break
            to_add.append(obj)
        else:
            raise InterpretationError("Exhausted the stack while searching for a MarkObject!")
        if not interpreter.stack:
            raise ValueError("Stack was empty; expected a pyset")
        pyset = interpreter.stack[-1]
        if not isinstance(pyset, ast.Set):
            raise ValueError(
                f"{pyset!r} was expected to be a set-like object with an `add` function"
            )
        # Check for cyclic references - sets cannot contain themselves (unhashable)
        for elem in to_add:
            if elem is pyset:
                raise InterpretationError("Set cannot contain itself (unhashable type)")
        pyset.elts.extend(reversed(to_add))


class Reduce(Opcode):
    name = "REDUCE"

    def run(self, interpreter: Interpreter):
        args = interpreter.stack.pop()
        func = interpreter.stack.pop()
        if isinstance(args, ast.Tuple):
            call = ast.Call(func, list(args.elts), [])
        else:
            call = ast.Call(func, [ast.Starred(args)], [])
        # Any call to reduce can have global side effects, since it runs arbitrary Python code.
        # However, if we just save it to the stack, then it might not make it to the final AST
        # unless the stack value is actually used. So save the result to a temp variable, and then
        # put that on the stack:
        var_name = interpreter.new_variable(call)
        interpreter.stack.append(ast.Name(var_name, ast.Load()))


class Mark(Opcode):
    name = "MARK"

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(MarkObject())


class Pop(Opcode):
    name = "POP"

    def run(self, interpreter: Interpreter):
        interpreter.stack.pop()


class PopMark(Opcode):
    name = "POP_MARK"

    def run(self, interpreter: Interpreter):
        objs = []
        while interpreter.stack:
            obj = interpreter.stack.pop()
            if isinstance(obj, MarkObject):
                break
            objs.append(obj)
        else:
            raise InterpretationError("Exhausted the stack while searching for a MarkObject!")
        return objs


class Obj(Opcode):
    name = "OBJ"

    def run(self, interpreter: Interpreter):
        args = []
        while interpreter.stack:
            arg = interpreter.stack.pop()
            if isinstance(arg, MarkObject):
                break
            args.insert(0, arg)
        else:
            raise InterpretationError("Exhausted the stack while searching for a MarkObject!")
        kls = args.pop(0)
        if args or hasattr(kls, "__getinitargs__") or not isinstance(kls, type):
            call = ast.Call(kls, args, [])
        else:
            # No args and kls is a plain type: CPython does kls.__new__(kls)
            call = ast.Call(ast.Attribute(kls, "__new__", ast.Load()), [kls], [])
        # OBJ calls can have global side effects, just like REDUCE.
        # Persist the call to the AST via new_variable() so it remains visible
        # to safety analysis even if the stack value is discarded by POP.
        var_name = interpreter.new_variable(call)
        interpreter.stack.append(ast.Name(var_name, ast.Load()))


class ShortBinUnicode(DynamicLength, ConstantOpcode):
    name = "SHORT_BINUNICODE"
    priority = 5000
    length_bytes = 1

    @classmethod
    def validate(cls, obj: str) -> bytes:
        if not isinstance(obj, str):
            raise ValueError(f"obj must be of type str, not {obj!r}")
        return super().validate(obj.encode("utf-8"))

    def encode_body(self) -> bytes:
        text = self.arg
        if isinstance(text, str):
            text = text.encode("utf-8")
        return text


class BinUnicode(ShortBinUnicode):
    name = "BINUNICODE"
    priority = ShortBinUnicode.priority + 1
    length_bytes = 4


class BinUnicode8(BinUnicode):
    name = "BINUNICODE8"
    priority = BinUnicode.priority + 1
    length_bytes = 8


class Unicode(ConstantOpcode):
    name = "UNICODE"
    priority = BinUnicode8.priority + 1

    @classmethod
    def validate(cls, obj: str) -> bytes:
        if not isinstance(obj, str):
            raise ValueError(f"{cls.__name__}.new expects a str object, not {obj!r}")
        return obj.encode("utf-8")

    def encode_body(self) -> bytes:
        return raw_unicode_escape(self.arg).encode("utf-8")


class String(ConstantOpcode):
    name = "STRING"
    priority = Unicode.priority + 1

    def encode_body(self) -> bytes:
        return repr(self.arg).encode("utf-8")

    @classmethod
    def validate(cls, obj):
        if not isinstance(obj, str):
            raise ValueError(f"String must be instantiated from a str, not {obj!r}")
        return obj


class NewObj(Opcode):
    name = "NEWOBJ"

    def run(self, interpreter: Interpreter):
        args = interpreter.stack.pop()
        class_type = interpreter.stack.pop()
        # CPython's NEWOBJ calls cls.__new__(cls, *args), not cls(*args).
        # __new__ is allocation only (no __init__ side effects).
        func = ast.Attribute(class_type, "__new__", ast.Load())
        if isinstance(args, ast.Tuple):
            call = ast.Call(func, [class_type, *list(args.elts)], [])
        else:
            call = ast.Call(func, [class_type, ast.Starred(args)], [])
        # Persist the call to the AST via new_variable() so it remains visible
        # to safety analysis even if the stack value is discarded by POP.
        var_name = interpreter.new_variable(call)
        interpreter.stack.append(ast.Name(var_name, ast.Load()))


class NewObjEx(Opcode):
    name = "NEWOBJ_EX"

    def run(self, interpreter: Interpreter):
        kwargs = interpreter.stack.pop()
        args = interpreter.stack.pop()
        class_type = interpreter.stack.pop()
        # CPython's NEWOBJ_EX calls cls.__new__(cls, *args, **kwargs), not cls(*args, **kwargs).
        func = ast.Attribute(class_type, "__new__", ast.Load())
        if isinstance(args, ast.Tuple):
            call = ast.Call(func, [class_type, *list(args.elts)], kwargs)
        else:
            call = ast.Call(func, [class_type, ast.Starred(args)], kwargs)
        # Persist the call to the AST via new_variable() so it remains visible
        # to safety analysis even if the stack value is discarded by POP.
        var_name = interpreter.new_variable(call)
        interpreter.stack.append(ast.Name(var_name, ast.Load()))


class BinPersId(Opcode):
    name = "BINPERSID"

    def run(self, interpreter: Interpreter):
        pid = interpreter.stack.pop()
        interpreter.stack.append(
            ast.Call(
                ast.Attribute(ast.Name("UNPICKLER", ast.Load()), "persistent_load"),
                [pid],
                [],
            )
        )


class PersId(Opcode):
    name = "PERSID"


class NoneOpcode(Opcode):
    name = "NONE"

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(make_constant(None))


class NewTrue(Opcode):
    name = "NEWTRUE"

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(make_constant(True))


class NewFalse(Opcode):
    name = "NEWFALSE"

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(make_constant(False))


class Tuple(StackSliceOpcode):
    name = "TUPLE"

    def run(self, interpreter: Interpreter, stack_slice: List[ast.expr]):
        interpreter.stack.append(ast.Tuple(tuple(stack_slice), ast.Load()))


class Build(Opcode):
    name = "BUILD"

    def run(self, interpreter: Interpreter):
        argument = interpreter.stack.pop()
        obj = interpreter.stack.pop()
        obj_name = interpreter.new_variable(obj)
        interpreter.module_body.append(
            ast.Expr(
                ast.Call(
                    ast.Attribute(ast.Name(obj_name, ast.Load()), "__setstate__"),
                    [argument],
                    [],
                )
            )
        )
        interpreter.stack.append(ast.Name(obj_name, ast.Load()))


class BinGet(Opcode):
    name = "BINGET"

    def run(self, interpreter: Interpreter):
        interpreter.track_get()
        if self.arg not in interpreter.memory:
            sys.stderr.write(
                f"Warning: malformed pickle file. BINGET references non-existent memo key {self.arg}; "
                f"using placeholder None value to continue analysis\n"
            )
            interpreter.stack.append(ast.Constant(value=None))
        else:
            interpreter.stack.append(interpreter.memory[self.arg])

    def encode_body(self):
        assert self.arg <= 255, "BINGET only supports 1-byte memo indexing"
        return bytes([self.arg])


class LongBinGet(Opcode):
    name = "LONG_BINGET"

    def run(self, interpreter: Interpreter):
        interpreter.track_get()
        if self.arg not in interpreter.memory:
            sys.stderr.write(
                f"Warning: malformed pickle file. LONG_BINGET references non-existent memo key {self.arg}; "
                f"using placeholder None value to continue analysis\n"
            )
            interpreter.stack.append(ast.Constant(value=None))
        else:
            interpreter.stack.append(interpreter.memory[self.arg])


class Get(Opcode):
    name = "GET"

    @property
    def memo_id(self) -> int:
        return int(self.arg)

    def run(self, interpreter: Interpreter):
        interpreter.track_get()
        if self.memo_id not in interpreter.memory:
            sys.stderr.write(
                f"Warning: malformed pickle file. BINGET references non-existent memo key {self.memo_id}; "
                f"using placeholder None value to continue analysis\n"
            )
            interpreter.stack.append(ast.Constant(value=None))
        else:
            interpreter.stack.append(interpreter.memory[self.memo_id])

    def encode_body(self) -> bytes:
        return f"{self.memo_id}\n".encode()

    @staticmethod
    def create(memo_id: int) -> Get:
        return Get(f"{memo_id}\n".encode())


class SetItems(StackSliceOpcode):
    name = "SETITEMS"

    def run(self, interpreter: Interpreter, stack_slice: list[ast.expr]):
        pydict = interpreter.stack.pop()
        update_dict_keys = []
        update_dict_values = []
        for key, value in zip(stack_slice[::2], stack_slice[1::2], strict=False):
            # Check for cyclic references
            if key is pydict:
                raise InterpretationError("Dict cannot use itself as key (unhashable type)")
            if value is pydict:
                value = ast.Set(elts=[ast.Constant(value=...)])
                interpreter._has_cycle = True
            update_dict_keys.append(key)
            update_dict_values.append(value)
        if isinstance(pydict, ast.Dict) and not pydict.keys:
            # the dict is empty, so add a new one
            interpreter.stack.append(ast.Dict(keys=update_dict_keys, values=update_dict_values))
        else:
            dict_name = interpreter.new_variable(pydict)
            update_dict = ast.Dict(keys=update_dict_keys, values=update_dict_values)
            interpreter.module_body.append(
                ast.Expr(
                    ast.Call(
                        ast.Attribute(ast.Name(dict_name, ast.Load()), "update"),
                        [update_dict],
                        [],
                    )
                )
            )
            interpreter.stack.append(ast.Name(dict_name, ast.Load()))


class SetItem(Opcode):
    name = "SETITEM"

    def run(self, interpreter: Interpreter):
        value = interpreter.stack.pop()
        key = interpreter.stack.pop()
        pydict = interpreter.stack.pop()
        # Check for cyclic references
        if key is pydict:
            raise InterpretationError("Dict cannot use itself as key (unhashable type)")
        if value is pydict:
            value = ast.Set(elts=[ast.Constant(value=...)])
            interpreter._has_cycle = True
        if isinstance(pydict, ast.Dict) and not pydict.keys:
            # the dict is empty, so add a new one
            interpreter.stack.append(ast.Dict(keys=[key], values=[value]))
        else:
            dict_name = interpreter.new_variable(pydict)
            assignment = ast.Assign(
                [ast.Subscript(ast.Name(dict_name, ast.Load()), key, ast.Store())],
                value,
            )
            interpreter.module_body.append(assignment)
            interpreter.stack.append(ast.Name(dict_name, ast.Load()))


class Stop(Opcode):
    name = "STOP"

    def run(self, interpreter: Interpreter):
        interpreter.new_variable(interpreter.stack.pop(), name=interpreter.result_variable)
        interpreter.stop()


class Frame(NoOp):
    name = "FRAME"


class BinInt1(ConstantInt):
    name = "BININT1"
    priority = 100
    num_bytes = 1


class BinInt2(BinInt1):
    name = "BININT2"
    priority = BinInt1.priority + 1
    num_bytes = 2


class BinInt(ConstantInt):
    name = "BININT"
    priority = BinInt2.priority + 1
    num_bytes = 4
    signed = True


class EmptySet(Opcode):
    name = "EMPTY_SET"

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(ast.Set([]))


class EmptyList(Opcode):
    name = "EMPTY_LIST"

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(ast.List([], ast.Load()))


class EmptyDict(Opcode):
    name = "EMPTY_DICT"

    def run(self, interpreter: Interpreter):
        interpreter.stack.append(ast.Dict(keys=[], values=[]))


class Memoize(Opcode):
    name = "MEMOIZE"

    def run(self, interpreter: Interpreter):
        interpreter.track_put()
        interpreter.memory[len(interpreter.memory)] = interpreter.stack[-1]


class Append(Opcode):
    name = "APPEND"

    def run(self, interpreter: Interpreter):
        value = interpreter.stack.pop()
        list_obj = interpreter.stack[-1]
        if isinstance(list_obj, ast.List):
            if value is list_obj:
                value = ast.List(elts=[ast.Constant(value=...)], ctx=ast.Load())
                interpreter._has_cycle = True
            list_obj.elts.append(value)
        else:
            raise ValueError(f"Expected a list on the stack, but instead found {list_obj!r}")


class Appends(StackSliceOpcode):
    name = "APPENDS"

    def run(self, interpreter: Interpreter, stack_slice: list[ast.expr]):
        list_obj = interpreter.stack[-1]
        if isinstance(list_obj, ast.List):
            for i, elem in enumerate(stack_slice):
                if elem is list_obj:
                    stack_slice[i] = ast.List(elts=[ast.Constant(value=...)], ctx=ast.Load())
                    interpreter._has_cycle = True
            list_obj.elts.extend(stack_slice)
        else:
            raise ValueError(f"Expected a list on the stack, but instead found {list_obj!r}")


class BinFloat(ConstantOpcode):
    name = "BINFLOAT"
    priority = BinInt1.priority * 2

    @classmethod
    def validate(cls, obj):
        if not isinstance(obj, float):
            raise ValueError(f"{cls.__name__} expects a float, but received {obj!r}")
        return obj


class ShortBinBytes(DynamicLength, ConstantOpcode):
    name = "SHORT_BINBYTES"
    priority = Unicode.priority + 1
    length_bytes = 1

    def encode_body(self) -> bytes:
        return self.arg

    @classmethod
    def validate(cls, obj):
        if not isinstance(obj, bytes):
            raise ValueError(
                f"{cls.__name__} must be instantiated with an object of type bytes, not {obj!r}"
            )
        return super().validate(obj)


class ShortBinString(DynamicLength, ConstantOpcode):
    name = "SHORT_BINSTRING"
    priority = Unicode.priority + 1
    length_bytes = 1

    def encode_body(self) -> bytes:
        return repr(self.arg).encode("utf-8")

    @classmethod
    def validate(cls, obj):
        if not isinstance(obj, str):
            raise ValueError(f"String must be instantiated from a str, not {obj!r}")
        return obj


class BinString(DynamicLength, ConstantOpcode):
    name = "BINSTRING"
    priority = ShortBinBytes.priority + 1
    length_bytes = 4
    signed = True

    def encode_body(self) -> bytes:
        return repr(self.arg).encode("utf-8")

    @classmethod
    def validate(cls, obj):
        if not isinstance(obj, str):
            raise ValueError(f"String must be instantiated from a str, not {obj!r}")
        return super().validate(obj)


class BinBytes(ShortBinBytes):
    name = "BINBYTES"
    priority = ShortBinBytes.priority + 1
    length_bytes = 4


class BinBytes8(BinBytes):
    name = "BINBYTES8"
    priority = BinBytes.priority + 1
    length_bytes = 8


class Long1(ConstantInt):
    name = "LONG1"
    num_bytes = 1
    signed = True
    priority = BinInt.priority + 1


class Long4(ConstantInt):
    name = "LONG4"
    num_bytes = 4
    signed = True
    priority = Long1.priority + 1


class Int(ConstantOpcode):
    name = "INT"
    priority = Long4.priority + 1

    def encode_body(self) -> bytes:
        return f"{int(self.arg)}\n".encode()

    @classmethod
    def validate(cls, obj):
        _ = int(obj)
        return obj


class Long(Int):
    name = "LONG"
    priority = Int.priority + 1


class Dict(Opcode):
    name = "DICT"

    def run(self, interpreter: Interpreter):
        i = 0
        keys = []
        values = []

        while interpreter.stack:
            obj = interpreter.stack.pop()
            if isinstance(obj, MarkObject):
                break
            if i == 0:
                values.append(obj)
            elif i == 1:
                keys.append(obj)
            i = (i + 1) % 2
        else:
            raise InterpretationError("Exhausted the stack while searching for a MarkObject!")

        if len(keys) != len(values):
            raise ValueError(
                f"Number of keys ({len(keys)}) and values ({len(values)}) for DICT do not match"
            )

        interpreter.stack.append(ast.Dict(keys=reversed(keys), values=reversed(values)))


PickledSequence = Sequence[Pickled]


class StackedPickle(PickledSequence):
    def __init__(self, pickled: Iterable[Pickled]):
        self.pickled: tuple[Pickled, ...] = tuple(pickled)

    def __getitem__(self, index: int) -> Pickled:
        return self.pickled[index]

    def __len__(self) -> int:
        return len(self.pickled)

    @staticmethod
    def load(pickled: Buffer | BinaryIO, fail_on_decode_error: bool = True) -> StackedPickle:
        pickled = Pickled.make_stream(pickled)
        pickles: List[Pickled] = []
        while True:
            try:
                p = Pickled.load(pickled, fail_on_decode_error=fail_on_decode_error)
                if len(p) == 0:
                    break
                pickles.append(p)
            except EmptyPickleError:
                break
        if not pickles:
            raise EmptyPickleError("No pickle files detected")
        return StackedPickle(pickles)


class List(Opcode):
    name = "LIST"

    def run(self, interpreter: Interpreter):
        objs = []
        while interpreter.stack:
            obj = interpreter.stack.pop()
            if isinstance(obj, MarkObject):
                break
            objs.append(obj)
        else:
            raise InterpretationError("Exhausted the stack while searching for a MarkObject!")

        interpreter.stack.append(ast.List(elts=objs[::-1], ctx=ast.Load()))


class FrozenSet(Opcode):
    name = "FROZENSET"

    def run(self, interpreter: Interpreter):
        objs = []
        while interpreter.stack:
            obj = interpreter.stack.pop()
            if isinstance(obj, MarkObject):
                break
            objs.append(obj)
        else:
            raise InterpretationError("Exhausted the stack while searching for a MarkObject!")

        interpreter.stack.append(ast.Constant(ast.Set(elts=objs[::-1])))


class Dup(Opcode):
    name = "DUP"

    def run(self, interpreter: Interpreter):
        stack_len = len(interpreter.stack)
        if stack_len == 0:
            raise IndexError(
                f"Opcode {self.opcode!s} \
                attempted to duplicate the topmost entry on the stack, but it is empty"
            )
        obj = interpreter.stack[stack_len - 1]
        interpreter.stack.append(obj)



import warnings

from fickling.fickle import *  # noqa: F403

warnings.warn(
    "As of version 0.0.8, `fickling.pickle` has been renamed `fickling.fickle`. "
    "The `fickling.pickle` module will be removed in a future release!",
    DeprecationWarning,
    stacklevel=2,
)



import pickle
import warnings
import zipfile
from io import BytesIO
from pathlib import PurePosixPath

from fickling.analysis import AnalysisResults, Severity, check_safety
from fickling.exception import UnsafeFileError
from fickling.fickle import Pickled, PickleDecodeError, StackedPickle


class RelaxedZipFile(zipfile.ZipFile):
    """A ZipFile subclass that ignores CRC validation errors.

    Matches PyTorch's lenient ZIP parsing behavior. Uses CPython's
    internal _expected_crc attribute on ZipExtFile — guarded by hasattr
    so it degrades to standard CRC-checked behavior if the attribute
    is renamed in a future Python version.
    """

    def open(self, name, mode="r", pwd=None, *, force_zip64=False):
        """Open a member with CRC validation disabled."""
        f = super().open(name, mode, pwd, force_zip64=force_zip64)
        if hasattr(f, "_expected_crc"):
            f._expected_crc = None
        else:
            warnings.warn(
                "RelaxedZipFile: _expected_crc not found on ZipExtFile. "
                "CRC validation is still enabled. This may cause failures "
                "scanning PyTorch model files with CRC mismatches.",
                RuntimeWarning,
                stacklevel=2,
            )
        return f


# Save the original pickle.loads before any hooks can replace it.
# loader.load() must use the real pickle.loads for final deserialization,
# otherwise hooking pickle.loads causes infinite recursion.
_original_pickle_loads = pickle.loads


def load(
    file,
    max_acceptable_severity=Severity.LIKELY_SAFE,
    print_results=False,
    json_output_path=None,
    *args,
    **kwargs,
):
    """Exposed as fickling.load()"""
    pickled_data = Pickled.load(file, fail_on_decode_error=False)
    result = check_safety(pickled=pickled_data, json_output_path=json_output_path)
    if result.severity <= max_acceptable_severity and not pickled_data.has_invalid_opcode:
        # We don't do pickle.load(file) because it could allow for a race
        # condition where the file we check is not the same that gets
        # loaded after the analysis.
        return _original_pickle_loads(pickled_data.dumps(), *args, **kwargs)
    if pickled_data.has_invalid_opcode:
        raise UnsafeFileError(
            file,
            "This file contains an invalid opcode sequence. It is "
            "either corrupted or maliciously attempting to bypass "
            "pickle analysis tools",
        )
    raise UnsafeFileError(file, result.to_dict())


def loads(
    data,
    max_acceptable_severity=Severity.LIKELY_SAFE,
    print_results=False,
    json_output_path=None,
    *args,
    **kwargs,
):
    """Exposed as fickling.loads()

    Safely loads a pickle from bytes data after performing security analysis.

    Args:
        data: Bytes data containing the pickled object
        max_acceptable_severity: Maximum acceptable severity level (default: LIKELY_SAFE)
        print_results: Whether to print analysis results (default: False)
        json_output_path: Optional path to write JSON analysis results
        *args: Additional arguments passed to pickle.loads()
        **kwargs: Additional keyword arguments passed to pickle.loads()

    Returns:
        The unpickled object

    Raises:
        UnsafeFileError: If the pickle data is unsafe or contains invalid opcodes
    """
    # Convert bytes to file-like object
    file = BytesIO(data)

    # Reuse existing load() function which already handles all the logic
    return load(
        file,
        *args,
        max_acceptable_severity=max_acceptable_severity,
        print_results=print_results,
        json_output_path=json_output_path,
        **kwargs,
    )


class ScanResult:
    """Result of scanning a file or archive for malicious pickle content."""

    def __init__(
        self,
        filepath: str,
        severity: Severity,
        results: list[AnalysisResults],
        errors: list[str],
    ):
        self.filepath = filepath
        self.severity = severity
        self.results = results
        self.errors = errors

    @property
    def is_safe(self) -> bool:
        return self.severity <= Severity.LIKELY_SAFE

    def __bool__(self) -> bool:
        return self.is_safe and not self.errors

    def __repr__(self) -> str:
        return (
            f"ScanResult(filepath={self.filepath!r}, "
            f"severity={self.severity.name}, "
            f"results={len(self.results)}, errors={len(self.errors)})"
        )


def scan_file(
    filepath: str,
    graceful: bool = False,
    json_output_path: str | None = None,
) -> ScanResult:
    """Scan a file for malicious pickle content.

    Args:
        filepath: Path to the file to scan
        graceful: If True, continue on parse errors and report them.
                  If False, raise exceptions on parse errors.
        json_output_path: Optional path to write JSON analysis results

    Returns:
        ScanResult with severity, results list, and errors list
    """
    try:
        with open(filepath, "rb") as f:
            data = f.read()
    except OSError as e:
        if graceful:
            return ScanResult(
                filepath=filepath,
                severity=Severity.SUSPICIOUS,
                results=[],
                errors=[f"File error ({type(e).__name__}): {e!s}"],
            )
        raise
    return _scan_bytes(filepath, data, graceful, json_output_path)


def scan_zip_archive(
    filepath: str,
    graceful: bool = False,
    json_output_path: str | None = None,
) -> dict[str, ScanResult]:
    """Scan a ZIP archive for malicious raw pickle content.

    Scans each file within the archive that has a raw pickle-related extension
    (.pkl, .pickle, .bin). Only ZIP archives are supported;
    for TAR or 7z archives, see fickling.polyglot.

    Args:
        filepath: Path to the ZIP archive to scan
        graceful: If True, continue on parse errors and report them
        json_output_path: Optional path to write JSON analysis results

    Returns:
        Dict mapping archive member names to their ScanResults
    """
    results: dict[str, ScanResult] = {}
    pickle_extensions = {".pkl", ".pickle", ".bin"}

    try:
        with RelaxedZipFile(filepath, "r") as archive:
            for info in archive.infolist():
                if info.is_dir():
                    continue
                ext = PurePosixPath(info.filename).suffix.lower()
                if ext not in pickle_extensions:
                    continue

                try:
                    data = archive.read(info)
                except (zipfile.BadZipFile, OSError) as e:
                    if graceful:
                        results[info.filename] = ScanResult(
                            filepath=info.filename,
                            severity=Severity.SUSPICIOUS,
                            results=[],
                            errors=[f"Read error ({type(e).__name__}): {e!s}"],
                        )
                        continue
                    raise

                results[info.filename] = _scan_bytes(
                    info.filename, data, graceful, json_output_path
                )
    except (zipfile.BadZipFile, OSError) as e:
        if not graceful:
            raise
        results["<archive>"] = ScanResult(
            filepath=filepath,
            severity=Severity.SUSPICIOUS,
            results=[],
            errors=[f"Archive error ({type(e).__name__}): {e!s}"],
        )

    return results


def _scan_bytes(
    name: str,
    data: bytes,
    graceful: bool,
    json_output_path: str | None,
) -> ScanResult:
    """Scan bytes data for malicious pickle content."""
    results: list[AnalysisResults] = []
    errors: list[str] = []
    overall_severity = Severity.LIKELY_SAFE

    try:
        stacked = StackedPickle.load(BytesIO(data), fail_on_decode_error=not graceful)
        for pickled in stacked:
            try:
                result = check_safety(pickled, json_output_path=json_output_path)
                results.append(result)
                if result.severity > overall_severity:
                    overall_severity = result.severity
            except Exception as e:
                if graceful:
                    errors.append(f"Analysis error ({type(e).__name__}): {e!s}")
                    overall_severity = max(overall_severity, Severity.LIKELY_UNSAFE)
                else:
                    raise
    except PickleDecodeError as e:
        if graceful:
            errors.append(f"Parse error ({type(e).__name__}): {e!s}")
            overall_severity = max(overall_severity, Severity.SUSPICIOUS)
        else:
            raise
    except Exception as e:
        if graceful:
            errors.append(f"Unexpected error ({type(e).__name__}): {e!s}")
            overall_severity = max(overall_severity, Severity.LIKELY_UNSAFE)
        else:
            raise

    return ScanResult(
        filepath=name,
        severity=overall_severity,
        results=results,
        errors=errors,
    )



# ClamAV-compatible exit codes for CI/CD integration
EXIT_CLEAN = 0  # No issues found
EXIT_UNSAFE = 1  # Potentially malicious content detected
EXIT_ERROR = 2  # Scan error (parse failure, file not found, etc.)
